"""ProtocolEngine-based Labware core implementations."""

from typing import List, Optional, cast, Dict

from opentrons_shared_data.labware.types import (
    LabwareParameters2 as LabwareParameters2Dict,
    LabwareParameters3 as LabwareParameters3Dict,
    LabwareDefinition as LabwareDefinitionDict,
)

from opentrons_shared_data.labware.labware_definition import LabwareRole

from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.errors import (
    LabwareNotOnDeckError,
    ModuleNotOnDeckError,
    LocationIsStagingSlotError,
)
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient
from opentrons.protocol_engine.types import (
    LabwareOffsetCreate,
    LabwareOffsetVector,
)
from opentrons.types import DeckSlotName, NozzleMapInterface, Point, StagingSlotName


from ..._liquid import Liquid
from ..labware import AbstractLabware, LabwareLoadParams

from .well import WellCore


_LabwareParametersDict = LabwareParameters2Dict | LabwareParameters3Dict


class LabwareCore(AbstractLabware[WellCore]):
    """Labware API core using a ProtocolEngine.

    Args:
        labware_id: ProtocolEngine ID of the loaded labware.
        engine_client: ProtocolEngine synchronous client.
    """

    def __init__(self, labware_id: str, engine_client: ProtocolEngineClient) -> None:
        self._labware_id = labware_id
        self._engine_client = engine_client

        labware_state = engine_client.state.labware
        self._definition = labware_state.get_definition(labware_id)
        self._user_display_name = labware_state.get_user_specified_display_name(
            labware_id
        )

    @property
    def labware_id(self) -> str:
        """The labware's unique ProtocolEngine ID."""
        return self._labware_id

    @property
    def highest_z(self) -> float:
        """The z-coordinate of the tallest single point anywhere on the labware."""
        return self._engine_client.state.geometry.get_labware_highest_z(
            self._labware_id
        )

    @property
    def load_name(self) -> str:
        """The API load name of the labware definition."""
        return self._definition.parameters.loadName

    def get_uri(self) -> str:
        """Get the URI string of the labware's definition.

        The URI is unique for a given namespace, load name, and definition version.
        """
        return self._engine_client.state.labware.get_definition_uri(self._labware_id)

    def get_load_params(self) -> LabwareLoadParams:
        return LabwareLoadParams(
            namespace=self._definition.namespace,
            load_name=self._definition.parameters.loadName,
            version=self._definition.version,
        )

    def get_display_name(self) -> str:
        """Get a display name for the labware, falling back to the definition."""
        return self._user_display_name or self._definition.metadata.displayName

    def get_user_display_name(self) -> Optional[str]:
        """Get the user-specified display name of the labware, if set."""
        return self._user_display_name

    def get_name(self) -> str:
        """Get the load name or the label of the labware specified by a user."""
        return self._user_display_name or self.load_name

    def get_definition(self) -> LabwareDefinitionDict:
        """Get the labware's definition as a plain dictionary."""
        return cast(
            LabwareDefinitionDict, self._definition.model_dump(exclude_none=True)
        )

    def get_parameters(self) -> _LabwareParametersDict:
        return cast(
            _LabwareParametersDict,
            self._definition.parameters.model_dump(exclude_none=True),
        )

    def get_quirks(self) -> List[str]:
        return self._definition.parameters.quirks or []

    def set_calibration(self, delta: Point) -> None:
        """Add a labware offset for this labware at its current location.

        This will override any previous labware offsets for this definition URI and location,
        even if the other labware offset was for a different specific labware instance.
        """
        offset_location = self._engine_client.state.geometry.get_offset_location(
            self._labware_id
        )
        if not offset_location:
            raise LabwareNotOnDeckError(
                message=f"Cannot set offset for {self.get_name()} as it is not currently in a deck slot.",
                details={"kind": "labware-not-in-slot"},
            )

        request = LabwareOffsetCreate.model_construct(
            definitionUri=self.get_uri(),
            locationSequence=offset_location,
            vector=LabwareOffsetVector(x=delta.x, y=delta.y, z=delta.z),
        )
        self._engine_client.add_labware_offset(request)
        self._engine_client.execute_command(
            cmd.ReloadLabwareParams(
                labwareId=self._labware_id,
            )
        )

    def get_calibrated_offset(self) -> Point:
        return self._engine_client.state.geometry.get_labware_position(self._labware_id)

    def is_tip_rack(self) -> bool:
        """Whether the labware is a tip rack."""
        return self._definition.parameters.isTiprack

    def is_adapter(self) -> bool:
        """Whether the labware is an adapter."""
        return LabwareRole.adapter in self._definition.allowedRoles

    def is_lid(self) -> bool:
        """Whether the labware is a lid."""
        return LabwareRole.lid in self._definition.allowedRoles

    def is_fixed_trash(self) -> bool:
        """Whether the labware is a fixed trash."""
        return self._engine_client.state.labware.is_fixed_trash(
            labware_id=self.labware_id
        )

    def get_tip_length(self) -> float:
        return self._engine_client.state.labware.get_tip_length(self._labware_id)

    def reset_tips(self) -> None:
        if self.is_tip_rack():
            self._engine_client.reset_tips(labware_id=self.labware_id)
        else:
            raise TypeError(f"{self.get_display_name()} is not a tip rack.")

    def get_next_tip(
        self,
        num_tips: int,
        starting_tip: Optional[WellCore],
        nozzle_map: Optional[NozzleMapInterface],
    ) -> Optional[str]:
        return self._engine_client.state.tips.get_next_tip(
            labware_id=self._labware_id,
            num_tips=num_tips,
            starting_tip_name=(
                starting_tip.get_name()
                if starting_tip and starting_tip.labware_id == self._labware_id
                else None
            ),
            nozzle_map=nozzle_map,
        )

    def get_well_columns(self) -> List[List[str]]:
        """Get the all well names, organized by column, from the labware's definition."""
        return self._definition.ordering

    def get_well_core(self, well_name: str) -> WellCore:
        """Create a well core interface to a well in this labware."""
        return WellCore(
            name=well_name,
            labware_id=self._labware_id,
            engine_client=self._engine_client,
        )

    def get_deck_slot(self) -> Optional[DeckSlotName]:
        """Get the deck slot the labware is in, if on deck."""
        try:
            ancestor = self._engine_client.state.geometry.get_ancestor_slot_name(
                self.labware_id
            )
            if isinstance(ancestor, StagingSlotName):
                # The only use case for get_deck_slot is with a legacy OT-2 function which resolves to a numerical deck slot, so we can ignore staging area slots for now
                return None
            return ancestor
        except (
            LabwareNotOnDeckError,
            ModuleNotOnDeckError,
            LocationIsStagingSlotError,
        ):
            return None

    def load_liquid(self, volumes: Dict[str, float], liquid: Liquid) -> None:
        """Load liquid into wells of the labware."""
        self._engine_client.execute_command(
            cmd.LoadLiquidParams(
                labwareId=self._labware_id, liquidId=liquid._id, volumeByWell=volumes
            )
        )

    def load_empty(self, wells: List[str]) -> None:
        """Mark wells of the labware as empty."""
        self._engine_client.execute_command(
            cmd.LoadLiquidParams(
                labwareId=self._labware_id,
                liquidId="EMPTY",
                volumeByWell={well: 0.0 for well in wells},
            )
        )



from enum import Enum
from dataclasses import dataclass, fields
from typing import List, Dict, Optional

from opentrons.drivers.command_builder import CommandBuilder


class GCODE(str, Enum):

    MOVE_TO = "G0"
    MOVE_TO_SWITCH = "G5"
    HOME_AXIS = "G28"
    STOP_MOTORS = "M0"
    ENABLE_MOTORS = "M17"
    GET_RESET_REASON = "M114"
    DEVICE_INFO = "M115"
    GET_LIMIT_SWITCH = "M119"
    GET_MOVE_PARAMS = "M120"
    GET_PLATFORM_SENSOR = "M121"
    GET_DOOR_SWITCH = "M122"
    GET_INSTALL_DETECTED = "M123"
    GET_STALLGUARD_THRESHOLD = "M911"
    GET_MOTOR_DRIVER_REGISTER = "M920"
    GET_TOF_SENSOR_STATUS = "M215"
    GET_TOF_DRIVER_REGISTER = "M222"
    GET_TOF_MEASUREMENT = "M226"
    ENABLE_TOF_SENSOR = "M224"
    MANAGE_TOF_MEASUREMENT = "M225"
    SET_TOF_CONFIGURATION = "M227"
    GET_TOF_CONFIGURATION = "M228"
    SET_LED = "M200"
    SET_SERIAL_NUMBER = "M996"
    SET_RUN_CURRENT = "M906"
    SET_IHOLD_CURRENT = "M907"
    SET_STALLGUARD = "M910"
    SET_MOTOR_DRIVER_REGISTER = "M921"
    SET_TOF_DRIVER_REGISTER = "M223"
    ENTER_BOOTLOADER = "dfu"

    def build_command(self) -> CommandBuilder:
        """Build command."""
        return CommandBuilder().add_gcode(self)


STACKER_VID = 0x483
STACKER_PID = 0xEF24
STACKER_FREQ = 115200


class HardwareRevision(Enum):
    """Hardware Revision."""

    NFF = "nff"
    EVT = "a1"
    DVT = "b1"


@dataclass
class StackerInfo:
    """Stacker Info."""

    fw: str
    hw: HardwareRevision
    sn: str
    rr: int = 0

    def to_dict(self) -> Dict[str, str]:
        """Build command."""
        return {
            "serial": self.sn,
            "version": self.fw,
            "model": self.hw.value,
            "reset_reason": str(self.rr),
        }


class StackerAxis(str, Enum):
    """Stacker Axis."""

    X = "X"
    Z = "Z"
    L = "L"


class TOFSensor(str, Enum):
    """Stacker TOF sensor."""

    X = "X"
    Z = "Z"


class LEDColor(Enum):
    """Stacker LED Color."""

    WHITE = 0
    RED = 1
    GREEN = 2
    BLUE = 3
    YELLOW = 4


class LEDPattern(Enum):
    """Stacker LED Pattern."""

    STATIC = 0
    FLASH = 1
    PULSE = 2
    CONFIRM = 3


class Direction(Enum):
    """Direction."""

    RETRACT = 0  # negative
    EXTEND = 1  # positive

    def __str__(self) -> str:
        """Convert to tag for clear logging."""
        return "negative" if self == Direction.RETRACT else "positive"

    def opposite(self) -> "Direction":
        """Get opposite direction."""
        return Direction.EXTEND if self == Direction.RETRACT else Direction.RETRACT

    def distance(self, distance: float) -> float:
        """Get signed distance, where retract direction is negative."""
        return distance * -1 if self == Direction.RETRACT else distance


class TOFSensorState(Enum):
    """TOF Sensor state."""

    DISABLED = 0
    INITIALIZING = 1
    IDLE = 2
    MEASURING = 3
    ERROR = 4


class TOFSensorMode(Enum):
    """The mode the sensor is in."""

    UNKNOWN = 0
    MEASURE = 0x03
    BOOTLOADER = 0x80


@dataclass
class LimitSwitchStatus:
    """Stacker Limit Switch Statuses."""

    XE: bool
    XR: bool
    ZE: bool
    ZR: bool
    LR: bool

    @classmethod
    def get_fields(cls) -> List[str]:
        """Get fields."""
        return [f.name for f in fields(cls)]

    def get(self, axis: StackerAxis, direction: Direction) -> bool:
        """Get limit switch status."""
        if axis == StackerAxis.X:
            return self.XE if direction == Direction.EXTEND else self.XR
        if axis == StackerAxis.Z:
            return self.ZE if direction == Direction.EXTEND else self.ZR
        if direction == Direction.EXTEND:
            raise ValueError("Latch does not have extent limit switch")
        return self.LR


@dataclass
class PlatformStatus:
    """Stacker Platform Statuses."""

    E: bool
    R: bool

    @classmethod
    def get_fields(cls) -> List[str]:
        """Get fields."""
        return [f.name for f in fields(cls)]

    def get(self, direction: Direction) -> bool:
        """Get platform status."""
        return self.E if direction == Direction.EXTEND else self.R

    def to_dict(self) -> Dict[str, bool]:
        """Dict of the data."""
        return {
            "extent": self.E,
            "retract": self.R,
        }


@dataclass
class TOFSensorStatus:
    """Stacker TOF sensor status."""

    sensor: TOFSensor
    state: TOFSensorState
    mode: TOFSensorMode
    ok: bool


@dataclass
class MoveParams:
    """Move Parameters."""

    max_speed: float
    acceleration: float
    max_speed_discont: float

    @classmethod
    def get_fields(cls) -> List[str]:
        """Get parsing fields."""
        return ["V", "A", "D"]

    def update(
        self,
        max_speed: Optional[float] = None,
        acceleration: Optional[float] = None,
        max_speed_discont: Optional[float] = None,
    ) -> "MoveParams":
        """Update the move parameters and return a new object."""
        return MoveParams(
            max_speed=max_speed if max_speed is not None else self.max_speed,
            acceleration=acceleration
            if acceleration is not None
            else self.acceleration,
            max_speed_discont=max_speed_discont
            if max_speed_discont is not None
            else self.max_speed_discont,
        )


@dataclass
class AxisParams:
    """Axis Parameters."""

    run_current: float
    hold_current: float
    move_params: MoveParams


@dataclass
class StallGuardParams:
    """StallGuard Parameters."""

    axis: StackerAxis
    enabled: bool
    threshold: int


class MoveResult(str, Enum):
    """The result of a move command."""

    NO_ERROR = "ok"
    STALL_ERROR = "stall"
    UNKNOWN_ERROR = "unknown"


class MeasurementKind(Enum):
    """The kind of measurement to request."""

    HISTOGRAM = 0


class SpadMapID(Enum):
    """The spad map id for the TOF sensor."""

    SPAD_MAP_ID_1 = 1
    # 3x3 macro 1 mode 33°x47° FoV off center
    SPAD_MAP_ID_2 = 2
    # 3x3 macro 2 mode 33°x47° FoV
    SPAD_MAP_ID_3 = 3
    # 3x3 wide mode 41°x52° FoV
    SPAD_MAP_ID_6 = 6
    # 3x3 mode 33°x32° FoV, checkerboard
    SPAD_MAP_ID_11 = 11
    # 3x3 mode 33°x32° FoV, inverted checkerboard
    SPAD_MAP_ID_12 = 12
    # User defined mode, single measurement mode
    SPAD_MAP_ID_14 = 14


class ActiveRange(Enum):
    """The active range for the TOF sensor."""

    NOT_SUPPORTED = 0
    SHORT_RANGE = 0x6E
    LONG_RANGE = 0x6F


@dataclass
class TOFMeasurement:
    """The start measurement data."""

    sensor: TOFSensor
    kind: MeasurementKind
    cancelled: bool
    total_bytes: int


@dataclass
class TOFMeasurementFrame:
    """Stacker TOF measurement frame."""

    sensor: TOFSensor
    frame_id: int
    data: bytes


@dataclass
class TOFMeasurementResult:
    """Stacker TOF measurement result."""

    sensor: TOFSensor
    kind: MeasurementKind
    bins: Dict[int, List[int]]


@dataclass
class TOFConfiguration:
    """Stacker TOF configuration."""

    sensor: TOFSensor
    spad_map_id: SpadMapID
    active_range: Optional[ActiveRange]
    kilo_iterations: Optional[int]
    report_period_ms: Optional[int]
    histogram_dump: Optional[bool]



"""Protocol engine utility types for model components."""
from dataclasses import dataclass

from pydantic import BaseModel


@dataclass(frozen=True)
class Dimensions:
    """Dimensions of an object in deck-space."""

    x: float
    y: float
    z: float


class Vec3f(BaseModel):
    """A 3D vector of floats."""

    x: float
    y: float
    z: float



"""Protocol engine errors module."""

from .exceptions import (
    ProtocolEngineError,
    UnexpectedProtocolError,
    FailedToLoadPipetteError,
    PipetteNotAttachedError,
    InvalidSpecificationForRobotTypeError,
    InvalidLoadPipetteSpecsError,
    TipNotAttachedError,
    PickUpTipTipNotAttachedError,
    TipAttachedError,
    CommandDoesNotExistError,
    UnsupportedLabwareForActionError,
    LabwareNotLoadedError,
    LabwareNotLoadedOnModuleError,
    LabwareNotLoadedOnLabwareError,
    LabwareNotOnDeckError,
    LiquidDoesNotExistError,
    LabwareDefinitionDoesNotExistError,
    LabwareCannotBeStackedError,
    LabwareCannotSitOnDeckError,
    LabwareIsInStackError,
    LabwareOffsetDoesNotExistError,
    LabwareIsNotTipRackError,
    LabwareIsTipRackError,
    LabwareIsAdapterError,
    TouchTipDisabledError,
    TouchTipIncompatibleArgumentsError,
    WellDoesNotExistError,
    PipetteNotLoadedError,
    ModuleNotLoadedError,
    ModuleNotOnDeckError,
    ModuleNotConnectedError,
    SlotDoesNotExistError,
    CutoutDoesNotExistError,
    FixtureDoesNotExistError,
    AddressableAreaDoesNotExistError,
    FixtureDoesNotProvideAreasError,
    AreaNotInDeckConfigurationError,
    IncompatibleAddressableAreaError,
    FailedToPlanMoveError,
    MustHomeError,
    RunStoppedError,
    SetupCommandNotAllowedError,
    FixitCommandNotAllowedError,
    ModuleNotAttachedError,
    ModuleAlreadyPresentError,
    WrongModuleTypeError,
    ThermocyclerNotOpenError,
    RobotDoorOpenError,
    PipetteMovementRestrictedByHeaterShakerError,
    HeaterShakerLabwareLatchNotOpenError,
    HeaterShakerLabwareLatchStatusUnknown,
    EngageHeightOutOfRangeError,
    NoTargetTemperatureSetError,
    InvalidTargetSpeedError,
    InvalidTargetTemperatureError,
    InvalidBlockVolumeError,
    InvalidHoldTimeError,
    InvalidWavelengthError,
    CannotPerformModuleAction,
    PauseNotAllowedError,
    ResumeFromRecoveryNotAllowedError,
    GripperNotAttachedError,
    CannotPerformGripperAction,
    HardwareNotSupportedError,
    LabwareMovementNotAllowedError,
    LabwareIsNotAllowedInLocationError,
    LocationIsOccupiedError,
    LocationNotAccessibleByPipetteError,
    LocationIsStagingSlotError,
    LocationIsLidDockSlotError,
    InvalidAxisForRobotType,
    NotSupportedOnRobotType,
    CommandNotAllowedError,
    InvalidLiquidHeightFound,
    LiquidHeightUnknownError,
    LiquidVolumeUnknownError,
    IncompleteLabwareDefinitionError,
    IncompleteWellDefinitionError,
    OperationLocationNotInWellError,
    InvalidDispenseVolumeError,
    StorageLimitReachedError,
    InvalidLiquidError,
    LiquidClassDoesNotExistError,
    LiquidClassRedefinitionError,
    OffsetLocationInvalidError,
    FlexStackerLabwarePoolNotYetDefinedError,
    FlexStackerNotLogicallyEmptyError,
    InvalidLabwarePositionError,
)

from .error_occurrence import ErrorOccurrence, ProtocolCommandFailedError

__all__ = [
    # exceptions
    "ProtocolEngineError",
    "UnexpectedProtocolError",
    "FailedToLoadPipetteError",
    "PipetteNotAttachedError",
    "InvalidSpecificationForRobotTypeError",
    "InvalidLoadPipetteSpecsError",
    "TipNotAttachedError",
    "PickUpTipTipNotAttachedError",
    "TipAttachedError",
    "CommandDoesNotExistError",
    "LabwareNotLoadedError",
    "LabwareNotLoadedOnModuleError",
    "LabwareNotLoadedOnLabwareError",
    "LabwareNotOnDeckError",
    "LiquidDoesNotExistError",
    "UnsupportedLabwareForActionError",
    "LabwareDefinitionDoesNotExistError",
    "LabwareCannotBeStackedError",
    "LabwareCannotSitOnDeckError",
    "LabwareIsInStackError",
    "LabwareOffsetDoesNotExistError",
    "LabwareIsNotTipRackError",
    "LabwareIsTipRackError",
    "LabwareIsAdapterError",
    "TouchTipDisabledError",
    "TouchTipIncompatibleArgumentsError",
    "WellDoesNotExistError",
    "PipetteNotLoadedError",
    "ModuleNotLoadedError",
    "ModuleNotOnDeckError",
    "ModuleNotConnectedError",
    "SlotDoesNotExistError",
    "CutoutDoesNotExistError",
    "FixtureDoesNotExistError",
    "AddressableAreaDoesNotExistError",
    "FixtureDoesNotProvideAreasError",
    "AreaNotInDeckConfigurationError",
    "IncompatibleAddressableAreaError",
    "FailedToPlanMoveError",
    "MustHomeError",
    "RunStoppedError",
    "SetupCommandNotAllowedError",
    "FixitCommandNotAllowedError",
    "ModuleNotAttachedError",
    "ModuleAlreadyPresentError",
    "WrongModuleTypeError",
    "ThermocyclerNotOpenError",
    "RobotDoorOpenError",
    "PipetteMovementRestrictedByHeaterShakerError",
    "HeaterShakerLabwareLatchNotOpenError",
    "HeaterShakerLabwareLatchStatusUnknown",
    "EngageHeightOutOfRangeError",
    "NoTargetTemperatureSetError",
    "InvalidTargetTemperatureError",
    "InvalidTargetSpeedError",
    "InvalidBlockVolumeError",
    "InvalidHoldTimeError",
    "InvalidLiquidError",
    "InvalidWavelengthError",
    "CannotPerformModuleAction",
    "ResumeFromRecoveryNotAllowedError",
    "PauseNotAllowedError",
    "ProtocolCommandFailedError",
    "GripperNotAttachedError",
    "CannotPerformGripperAction",
    "HardwareNotSupportedError",
    "LabwareMovementNotAllowedError",
    "LabwareIsNotAllowedInLocationError",
    "LocationIsOccupiedError",
    "LocationNotAccessibleByPipetteError",
    "LocationIsStagingSlotError",
    "LocationIsLidDockSlotError",
    "InvalidAxisForRobotType",
    "NotSupportedOnRobotType",
    "OffsetLocationInvalidError",
    "FlexStackerLabwarePoolNotYetDefinedError",
    "FlexStackerNotLogicallyEmptyError",
    "InvalidLabwarePositionError",
    # error occurrence models
    "ErrorOccurrence",
    "CommandNotAllowedError",
    "InvalidLiquidHeightFound",
    "LiquidHeightUnknownError",
    "LiquidVolumeUnknownError",
    "IncompleteLabwareDefinitionError",
    "IncompleteWellDefinitionError",
    "OperationLocationNotInWellError",
    "InvalidDispenseVolumeError",
    "StorageLimitReachedError",
    "LiquidClassDoesNotExistError",
    "LiquidClassRedefinitionError",
]



"""Command models for opening a gripper jaw."""

from __future__ import annotations
from typing import Literal, Type, Optional, Any, TYPE_CHECKING

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.resources import ensure_ot3_hardware

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ...state.state import StateView


CloseGripperJawCommandType = Literal["robot/closeGripperJaw"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class CloseGripperJawParams(BaseModel):
    """Payload required to close a gripper."""

    force: float | SkipJsonSchema[None] = Field(
        default=None,
        description="The force the gripper should use to hold the jaws, falls to default if none is provided.",
        json_schema_extra=_remove_default,
    )


class CloseGripperJawResult(BaseModel):
    """Result data from the execution of a CloseGripperJaw command."""

    pass


class CloseGripperJawImplementation(
    AbstractCommandImpl[CloseGripperJawParams, SuccessData[CloseGripperJawResult]]
):
    """CloseGripperJaw command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api
        self._state_view = state_view

    async def execute(
        self, params: CloseGripperJawParams
    ) -> SuccessData[CloseGripperJawResult]:
        """Release the gripper."""
        if self._state_view.config.use_virtual_gripper:
            return SuccessData(
                public=CloseGripperJawResult(),
            )
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        await ot3_hardware_api.grip(force_newtons=params.force)
        return SuccessData(
            public=CloseGripperJawResult(),
        )


class CloseGripperJaw(
    BaseCommand[CloseGripperJawParams, CloseGripperJawResult, ErrorOccurrence]
):
    """CloseGripperJaw command model."""

    commandType: CloseGripperJawCommandType = "robot/closeGripperJaw"
    params: CloseGripperJawParams
    result: Optional[CloseGripperJawResult] = None

    _ImplementationCls: Type[
        CloseGripperJawImplementation
    ] = CloseGripperJawImplementation


class CloseGripperJawCreate(BaseCommandCreate[CloseGripperJawParams]):
    """CloseGripperJaw command request model."""

    commandType: CloseGripperJawCommandType = "robot/closeGripperJaw"
    params: CloseGripperJawParams

    _CommandCls: Type[CloseGripperJaw] = CloseGripperJaw



from .abstract import AbstractFlexStackerDriver
from .driver import FlexStackerDriver, STACKER_MOTION_CONFIG, STALLGUARD_CONFIG
from .simulator import SimulatingDriver
from . import types as FlexStackerTypes
from . import utils as FlexStackerUtils

__all__ = [
    "AbstractFlexStackerDriver",
    "FlexStackerDriver",
    "SimulatingDriver",
    "FlexStackerTypes",
    "FlexStackerUtils",
    "STACKER_MOTION_CONFIG",
    "STALLGUARD_CONFIG",
]



from __future__ import annotations
from typing_extensions import Final
import enum


# TODO (tz, 5-18-23): think about a better name for it that would also work when we include staging area slots in the type.
class OffDeckType(enum.Enum):
    OFF_DECK = "off-deck"


OFF_DECK: Final = OffDeckType.OFF_DECK

# Set __doc__ manually as a workaround. When this docstring is written the normal way, right after
# the constant definition, Sphinx has trouble picking it up.
OFF_DECK.__doc__ = """\
A special location value, indicating that a labware is not currently on the robot's deck.

See :ref:`off-deck-location` for details on using ``OFF_DECK`` with :py:obj:`ProtocolContext.move_labware()`.
"""


class PlungerPositionTypes(enum.Enum):
    PLUNGER_TOP = "top"
    PLUNGER_BOTTOM = "bottom"
    PLUNGER_BLOWOUT = "blow_out"
    PLUNGER_DROPTIP = "drop_tip"


PLUNGER_TOP: Final = PlungerPositionTypes.PLUNGER_TOP
PLUNGER_BOTTOM: Final = PlungerPositionTypes.PLUNGER_BOTTOM
PLUNGER_BLOWOUT: Final = PlungerPositionTypes.PLUNGER_BLOWOUT
PLUNGER_DROPTIP: Final = PlungerPositionTypes.PLUNGER_DROPTIP


class PipetteActionTypes(enum.Enum):
    ASPIRATE_ACTION = "aspirate"
    DISPENSE_ACTION = "dispense"
    BLOWOUT_ACTION = "blowout"


ASPIRATE_ACTION: Final = PipetteActionTypes.ASPIRATE_ACTION
DISPENSE_ACTION: Final = PipetteActionTypes.DISPENSE_ACTION
BLOWOUT_ACTION: Final = PipetteActionTypes.BLOWOUT_ACTION



"""Executor for liquid class based complex commands."""
from __future__ import annotations

import logging
from copy import deepcopy
from enum import Enum
from typing import TYPE_CHECKING, Optional, Union, Literal
from dataclasses import dataclass, field

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    PositionReference,
    Coordinate,
    BlowoutLocation,
)

from opentrons.protocol_api._liquid_properties import (
    Submerge,
    TransferProperties,
    MixProperties,
    SingleDispenseProperties,
    MultiDispenseProperties,
    TouchTipProperties,
)
from opentrons.protocol_engine.errors import TouchTipDisabledError
from opentrons.types import Location, Point, Mount
from opentrons.protocols.advanced_control.transfers.transfer_liquid_utils import (
    LocationCheckDescriptors,
    check_current_volume_before_dispensing,
)
from opentrons.protocols.advanced_control.transfers import (
    transfer_liquid_utils as tx_utils,
)

if TYPE_CHECKING:
    from .well import WellCore
    from .instrument import InstrumentCore
    from ... import TrashBin, WasteChute

log = logging.getLogger(__name__)


AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP = 2


@dataclass
class LiquidAndAirGapPair:
    """Pairing of a liquid and air gap in a tip, with air gap below the liquid in a tip."""

    liquid: float = 0
    air_gap: float = 0


@dataclass
class TipState:
    """Carrier of the state of the pipette tip in use.

    Properties:
        last_liquid_and_air_gap_in_tip: The last liquid + air_gap combo in the tip.
            This will only include the existing liquid and air gap in the tip that
            an aspirate/ dispense interacts with. For example, the air gap from
            a previous step that needs to be removed, or the liquid from a previous
            aspirate that needs to be dispensed or the liquid that needs to be added to
            during a consolidation.
        ready_to_aspirate: Whether the pipette plunger is in a position that allows
            correct aspiration. The starting state for the pipette at initialization of
            `TransferComponentsExecutor`s should be ready_to_aspirate == True.
    """

    ready_to_aspirate: bool = True
    # TODO: maybe use the tip contents from engine state instead.
    last_liquid_and_air_gap_in_tip: LiquidAndAirGapPair = field(
        default_factory=LiquidAndAirGapPair
    )

    def append_liquid(self, volume: float) -> None:
        # Neither aspirate nor a dispense process should be adding liquid
        # when there is an air gap present.
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == 0
        ), "Air gap present in the tip."
        self.last_liquid_and_air_gap_in_tip.liquid += volume

    def delete_liquid(self, volume: float) -> None:
        # Neither aspirate nor a dispense process should be removing liquid
        # when there is an air gap present.
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == 0
        ), "Air gap present in the tip."
        self.last_liquid_and_air_gap_in_tip.liquid -= volume

    def append_air_gap(self, volume: float) -> None:
        # Neither aspirate nor a dispense process should be adding air gaps
        # when there is already an air gap present.
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == 0
        ), "Air gap already present in the tip."
        self.last_liquid_and_air_gap_in_tip.air_gap = volume

    def delete_air_gap(self, volume: float) -> None:
        assert (
            self.last_liquid_and_air_gap_in_tip.air_gap == volume
        ), "Last air gap volume doe not match the volume being removed"
        self.last_liquid_and_air_gap_in_tip.air_gap = 0

    def delete_last_air_gap_and_liquid(self) -> None:
        air_gap_in_tip = self.last_liquid_and_air_gap_in_tip.air_gap
        liquid_in_tip = self.last_liquid_and_air_gap_in_tip.liquid
        if air_gap_in_tip:
            self.delete_air_gap(air_gap_in_tip)
        if liquid_in_tip:
            self.delete_liquid(volume=liquid_in_tip)


class TransferType(Enum):
    ONE_TO_ONE = "one_to_one"
    MANY_TO_ONE = "many_to_one"
    ONE_TO_MANY = "one_to_many"


class TransferComponentsExecutor:
    def __init__(
        self,
        instrument_core: InstrumentCore,
        transfer_properties: TransferProperties,
        target_location: Union[Location, TrashBin, WasteChute],
        target_well: Optional[WellCore],
        tip_state: TipState,
        transfer_type: TransferType,
    ) -> None:
        """Create a TransferComponentsExecutor instance.

        One instance should be created to execute all the steps inside each of the
        liquid class' transfer components- aspirate, dispense and multi-dispense.
        The state of the TransferComponentsExecutor instance is expected to be valid
        only for the component it was created.

        For example, if we want to execute all the steps (submerge, dispense, retract, etc)
        related to the 'dispense' component of a liquid-class based transfer, the class
        will be used to initialize info about the dispense by assigning values
        to class attributes as follows-
        - target_location: the dispense location
        - target_well: the well associated with dispense location, will be None when the
                        target_location argument is a TrashBin or WasteChute
        - tip_state: the state of the tip before dispense component steps are executed
        - transfer_type: whether the dispense component is being called as a part of a
                        1-to-1 transfer or a consolidation or a distribution

        These attributes will remain the same throughout the component's execution,
        except `tip_state`, which will keep updating as fluids are handled.
        """
        self._instrument = instrument_core
        self._transfer_properties = transfer_properties
        self._target_location = target_location
        self._target_well = target_well
        self._tip_state: TipState = deepcopy(tip_state)  # don't modify caller's object
        self._transfer_type: TransferType = transfer_type

    @property
    def tip_state(self) -> TipState:
        """Return the tip state."""
        return self._tip_state

    def submerge(
        self,
        submerge_properties: Submerge,
        post_submerge_action: Literal["aspirate", "dispense"],
    ) -> None:
        """Execute submerge steps.

        1. move to position shown by positionReference + offset (should practically be a point outside/above the liquid).
        Should raise an error if this point is inside the liquid?
            For liquid meniscus this is easy to tell. Can’t be below meniscus
            For reference pos of anything else, do not allow submerge position to be below aspirate position
        2. move to aspirate/dispense position at desired speed
        3. delay

        If target location is a trash bin or waste chute, the pipette will move to the disposal location given,
        remove air gap and delay
        """
        submerge_start_location: Union[Location, TrashBin, WasteChute]
        if isinstance(self._target_location, Location):
            assert self._target_well is not None
            submerge_start_point = absolute_point_from_position_reference_and_offset(
                well=self._target_well,
                well_volume_difference=0,
                position_reference=submerge_properties.start_position.position_reference,
                offset=submerge_properties.start_position.offset,
                mount=self._instrument.get_mount(),
            )
            submerge_start_location = Location(
                point=submerge_start_point, labware=self._target_location.labware
            )
            tx_utils.raise_if_location_inside_liquid(
                location=submerge_start_location,
                well_core=self._target_well,
                location_check_descriptors=LocationCheckDescriptors(
                    location_type="submerge start",
                    pipetting_action=post_submerge_action,
                ),
                logger=log,
            )
        else:
            submerge_start_location = self._target_location

        self._instrument.move_to(
            location=submerge_start_location,
            well_core=self._target_well,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
        )
        self._remove_air_gap(location=submerge_start_location)
        if isinstance(self._target_location, Location):
            self._instrument.move_to(
                location=self._target_location,
                well_core=self._target_well,
                force_direct=True,
                minimum_z_height=None,
                speed=submerge_properties.speed,
            )

        if submerge_properties.delay.enabled and submerge_properties.delay.duration:
            self._instrument.delay(submerge_properties.delay.duration)

    def aspirate_and_wait(self, volume: float) -> None:
        """Aspirate according to aspirate properties and wait if enabled."""
        # TODO: handle volume correction
        assert (
            isinstance(self._target_location, Location)
            and self._target_well is not None
        )
        aspirate_props = self._transfer_properties.aspirate
        correction_volume = aspirate_props.correction_by_volume.get_for_volume(
            self._instrument.get_current_volume() + volume
        )
        self._instrument.aspirate(
            location=self._target_location,
            well_core=None,
            volume=volume,
            rate=1,
            flow_rate=aspirate_props.flow_rate_by_volume.get_for_volume(volume),
            in_place=True,
            correction_volume=correction_volume,
        )
        self._tip_state.append_liquid(volume)
        delay_props = aspirate_props.delay
        if delay_props.enabled and delay_props.duration:
            self._instrument.delay(delay_props.duration)

    def dispense_and_wait(
        self,
        dispense_properties: Union[SingleDispenseProperties, MultiDispenseProperties],
        volume: float,
        push_out_override: Optional[float],
    ) -> None:
        """Dispense according to dispense properties and wait if enabled."""
        current_vol = self._instrument.get_current_volume()
        check_current_volume_before_dispensing(
            current_volume=current_vol, dispense_volume=volume
        )
        correction_volume = dispense_properties.correction_by_volume.get_for_volume(
            current_vol - volume
        )
        self._instrument.dispense(
            location=self._target_location,
            well_core=None,
            volume=volume,
            rate=1,
            flow_rate=dispense_properties.flow_rate_by_volume.get_for_volume(volume),
            in_place=True,
            push_out=push_out_override,
            correction_volume=correction_volume,
        )
        if push_out_override:
            # If a push out was performed, we need to reset the plunger before we can aspirate again
            self._tip_state.ready_to_aspirate = False
        self._tip_state.delete_liquid(volume)
        dispense_delay = dispense_properties.delay
        if dispense_delay.enabled and dispense_delay.duration:
            self._instrument.delay(dispense_delay.duration)

    def mix(self, mix_properties: MixProperties, last_dispense_push_out: bool) -> None:
        """Execute mix steps.

        1. Use same flow rates and delays as aspirate and dispense
        2. Do [(aspirate + dispense) x repetitions] at the same position
        3. Do NOT push out at the end of dispense
        4. USE the delay property from aspirate & dispense during mix as well (flow rate and delay are coordinated with each other)
        5. Do not mix during consolidation
        NOTE: For most of our built-in definitions, we will keep _mix_ off because it is a very application specific thing.
        We should mention in our docs that users should adjust this property according to their application.
        """
        if not mix_properties.enabled or not isinstance(
            self._target_location, Location
        ):
            return
        # Assertion only for mypy purposes
        assert (
            mix_properties.repetitions is not None
            and mix_properties.volume is not None
            and self._target_well is not None
        )
        push_out_vol = (
            self._transfer_properties.dispense.push_out_by_volume.get_for_volume(
                mix_properties.volume
            )
        )
        for n in range(mix_properties.repetitions, 0, -1):
            self.aspirate_and_wait(volume=mix_properties.volume)
            self.dispense_and_wait(
                dispense_properties=self._transfer_properties.dispense,  # TODO: check that using single-dispense props during mix is correct
                volume=mix_properties.volume,
                push_out_override=push_out_vol
                if last_dispense_push_out is True and n == 1
                else 0,
            )

    def pre_wet(
        self,
        volume: float,
    ) -> None:
        """Do a pre-wet.

        - 1 combo of aspirate + dispense at the same flow rate as specified in asp & disp and the delays in asp & disp
        - Use the target volume/ volume we will be aspirating
        - No push out
        - No pre-wet for consolidation
        """
        if not self._transfer_properties.aspirate.pre_wet:
            return
        mix_props = MixProperties(_enabled=True, _repetitions=1, _volume=volume)
        self.mix(mix_properties=mix_props, last_dispense_push_out=False)

    def retract_after_aspiration(
        self, volume: float, add_air_gap: Optional[bool] = True
    ) -> None:
        """Execute post-aspiration retraction steps.

        1. Move TO the position reference+offset AT the specified speed
            Raise error if retract is below aspirate position or below the meniscus
        2. Delay
        3. Touch tip
            - Move to the Z offset position
            - Touch tip to the sides at the specified speed (tip moves back to the center as part of touch tip)
            - Return back to the retract position
        4. Air gap
            - If the retract location is at or above the safe location of
              AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP, then add the air gap at the
              retract location (where the pipette is already assumed to be).
            - If the retract location is below the safe location, then move to
              the safe location and then add the air gap.
            - Air gap volume depends on the amount of liquid in the pipette.
              So, if the total aspirated volume is 20, use the value for airGapByVolume[20]
              Flow rate = max(aspirateFlowRate, (airGapByVolume)/sec)
            - Use post-aspirate delay

        Args:
            volume: dispense volume
            add_air_gap: whether to add an air gap before moving away from the current well.
                         This value is True for all retractions, except when retracting
                         during a multi-dispense. Value of add_air_gap during multi-dispense
                         will depend on whether a conditioning volume is used.
        """
        assert (
            isinstance(self._target_location, Location)
            and self._target_well is not None
        )
        retract_props = self._transfer_properties.aspirate.retract
        retract_point = absolute_point_from_position_reference_and_offset(
            well=self._target_well,
            well_volume_difference=0,
            position_reference=retract_props.end_position.position_reference,
            offset=retract_props.end_position.offset,
            mount=self._instrument.get_mount(),
        )
        retract_location = Location(
            retract_point, labware=self._target_location.labware
        )
        tx_utils.raise_if_location_inside_liquid(
            location=retract_location,
            well_core=self._target_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="aspirate",
            ),
            logger=log,
        )
        self._instrument.move_to(
            location=retract_location,
            well_core=self._target_well,
            force_direct=True,
            minimum_z_height=None,
            speed=retract_props.speed,
        )
        retract_delay = retract_props.delay
        if retract_delay.enabled and retract_delay.duration:
            self._instrument.delay(retract_delay.duration)
        touch_tip_props = retract_props.touch_tip
        if touch_tip_props.enabled:
            assert (
                touch_tip_props.speed is not None
                and touch_tip_props.z_offset is not None
                and touch_tip_props.mm_from_edge is not None
            )
            self._instrument.touch_tip(
                location=retract_location,
                well_core=self._target_well,
                radius=1,
                z_offset=touch_tip_props.z_offset,
                speed=touch_tip_props.speed,
                mm_from_edge=touch_tip_props.mm_from_edge,
            )
            self._instrument.move_to(
                location=retract_location,
                well_core=self._target_well,
                force_direct=True,
                minimum_z_height=None,
                # Full speed because the tip will already be out of the liquid
                speed=None,
            )
        # For consolidate, we need to know the total amount that is in the pipette
        # since this may not be the first aspirate
        if self._transfer_type == TransferType.MANY_TO_ONE:
            volume_for_air_gap = self._instrument.get_current_volume()
        else:
            volume_for_air_gap = volume
        if add_air_gap:
            # If we need to add air gap, move to a safe location above the well if
            # the retract location is not already at or above this safe location
            if (
                retract_location.point.z
                < self._target_well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP).z
            ):
                self._instrument.move_to(
                    location=Location(
                        point=Point(
                            retract_location.point.x,
                            retract_location.point.y,
                            self._target_well.get_top(
                                AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP
                            ).z,
                        ),
                        labware=retract_location.labware,
                    ),
                    well_core=self._target_well,
                    force_direct=True,
                    minimum_z_height=None,
                    # Full speed because the tip will already be out of the liquid
                    speed=None,
                )
            self._add_air_gap(
                air_gap_volume=self._transfer_properties.aspirate.retract.air_gap_by_volume.get_for_volume(
                    volume_for_air_gap
                )
            )

    def retract_after_dispensing(
        self,
        trash_location: Union[Location, TrashBin, WasteChute],
        source_location: Optional[Location],
        source_well: Optional[WellCore],
        add_final_air_gap: bool,
    ) -> None:
        """Execute post-dispense retraction steps.
        1. Position ref+offset is the ending position. Move to this position using specified speed
        2. If blowout is enabled and “destination”
            - Do blow-out (at the retract position)
            - Leave plunger down
        3. Touch-tip
        4. If not ready-to-aspirate
            - Prepare-to-aspirate (at the retract position)
        5. Air-gap (at the retract position)
            - This air gap is for preventing any stray droplets from falling while moving the pipette.
                It will be performed out of caution even if we just did a blow_out and should *hypothetically*
                have no liquid left in the tip.
            - This air gap will be removed at the next aspirate.
                If this is the last step of the transfer, and we aren't dropping the tip off,
                then the air gap will be left as is(?).
        6. If blowout is “source” or “trash”
            - Move to location (top of Well)
            - Do blow-out (top of well)
            - Do touch-tip (?????) (only if it’s in a non-trash location)
            - Prepare-to-aspirate (top of well)
            - Do air-gap (top of well)
        7. If drop tip, move to drop tip location, drop tip

        If target location is a trash bin or waste chute, the retract movement step is skipped along with touch tip,
        even if it is enabled.
        """
        retract_props = self._transfer_properties.dispense.retract

        retract_location: Union[Location, TrashBin, WasteChute]
        if isinstance(self._target_location, Location):
            assert self._target_well is not None
            retract_point = absolute_point_from_position_reference_and_offset(
                well=self._target_well,
                well_volume_difference=0,
                position_reference=retract_props.end_position.position_reference,
                offset=retract_props.end_position.offset,
                mount=self._instrument.get_mount(),
            )
            retract_location = Location(
                retract_point, labware=self._target_location.labware
            )
            tx_utils.raise_if_location_inside_liquid(
                location=retract_location,
                well_core=self._target_well,
                location_check_descriptors=LocationCheckDescriptors(
                    location_type="retract end",
                    pipetting_action="dispense",
                ),
                logger=log,
            )
            self._instrument.move_to(
                location=retract_location,
                well_core=self._target_well,
                force_direct=True,
                minimum_z_height=None,
                speed=retract_props.speed,
            )
        else:
            retract_location = self._target_location

        # TODO should we delay here for a trash despite not having a "retract"?
        retract_delay = retract_props.delay
        if retract_delay.enabled and retract_delay.duration:
            self._instrument.delay(retract_delay.duration)

        blowout_props = retract_props.blowout
        if (
            blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ):
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            self._instrument.blow_out(
                location=retract_location,
                well_core=None,
                in_place=True,
            )
            self._tip_state.ready_to_aspirate = False
        is_final_air_gap = (
            blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ) or not blowout_props.enabled

        if is_final_air_gap and not add_final_air_gap:
            air_gap_volume = 0.0
        else:
            air_gap_volume = retract_props.air_gap_by_volume.get_for_volume(0)
        # Regardless of the blowout location, do touch tip and air gap
        # when leaving the dispense well. If this will be the final air gap, i.e,
        # we won't be moving to a Trash or a Source for Blowout after this air gap,
        # then skip the final air gap if we have been told to do so.
        self._do_touch_tip_and_air_gap_after_dispense(
            touch_tip_properties=retract_props.touch_tip,
            location=retract_location,
            well=self._target_well,
            air_gap_volume=air_gap_volume,
        )

        if (
            blowout_props.enabled
            and blowout_props.location != BlowoutLocation.DESTINATION
        ):
            # TODO: no-op touch tip if touch tip is enabled and blowout is in trash/ reservoir/ any labware with touch-tip disabled
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            touch_tip_and_air_gap_location: Union[Location, TrashBin, WasteChute]
            if blowout_props.location == BlowoutLocation.SOURCE:
                if source_location is None or source_well is None:
                    raise RuntimeError(
                        "Blowout location is 'source' but source location &/or well is not provided."
                    )
                # TODO: check if we should add a blowout location z-offset in liq class definition
                self._instrument.blow_out(
                    location=Location(
                        source_well.get_top(0), labware=source_location.labware
                    ),
                    well_core=source_well,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = Location(
                    source_well.get_top(0), labware=source_location.labware
                )
                touch_tip_and_air_gap_well = source_well
            else:
                self._instrument.blow_out(
                    location=trash_location,
                    well_core=None,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = trash_location
                touch_tip_and_air_gap_well = (
                    # We have already established that trash location of `Location` type
                    # has its `labware` as `Well` type.
                    trash_location.labware.as_well()._core  # type: ignore[assignment]
                    if isinstance(trash_location, Location)
                    else None
                )
            # A non-multi-dispense blowout will only have air and maybe droplets in the tip
            # since we only blowout after dispensing the full tip contents.
            # So delete the air gap from tip state
            last_air_gap = self._tip_state.last_liquid_and_air_gap_in_tip.air_gap
            self._tip_state.delete_air_gap(last_air_gap)
            self._tip_state.ready_to_aspirate = False

            air_gap_volume = (
                retract_props.air_gap_by_volume.get_for_volume(0)
                if add_final_air_gap
                else 0.0
            )
            # Do touch tip and air gap again after blowing out into source well or trash
            self._do_touch_tip_and_air_gap_after_dispense(
                touch_tip_properties=retract_props.touch_tip,
                location=touch_tip_and_air_gap_location,
                well=touch_tip_and_air_gap_well,
                air_gap_volume=air_gap_volume,
            )

    def retract_during_multi_dispensing(  # noqa: C901
        self,
        trash_location: Union[Location, TrashBin, WasteChute],
        source_location: Optional[Location],
        source_well: Optional[WellCore],
        conditioning_volume: float,
        add_final_air_gap: bool,
        is_last_retract: bool,
    ) -> None:
        """Execute post-dispense retraction steps when the dispense is a part of a multi-dispense.

        Args:
            trash_location: Location where we can drop tips or blowout, if set to do so
            source_location: Location where we can blowout, if set to do so
            source_well: Well where we can blowout, if set to do so
            conditioning_volume: Conditioning volume used for this multi-dispense. Can be 0
            add_final_air_gap: Whether we should add the final air gap of the step
            is_last_retract: Whether this is the last retract of the multi-dispense steps,
                i.e., this is part of the last dispense in the series of consecutive dispenses.
                This dispense might not be the last dispense of the entire distribution.

        This function is mostly similar to the single-dispense retract function except
        that it handles air gaps differently based on the disposal volume, conditioning volume
        and whether we are moving to another dispense or going back to the source.
        """
        assert (
            isinstance(self._target_location, Location)
            and self._target_well is not None
        )
        assert self._transfer_properties.multi_dispense is not None

        retract_props = self._transfer_properties.multi_dispense.retract
        retract_point = absolute_point_from_position_reference_and_offset(
            well=self._target_well,
            well_volume_difference=0,
            position_reference=retract_props.end_position.position_reference,
            offset=retract_props.end_position.offset,
            mount=self._instrument.get_mount(),
        )
        retract_location = Location(
            retract_point, labware=self._target_location.labware
        )
        tx_utils.raise_if_location_inside_liquid(
            location=retract_location,
            well_core=self._target_well,
            location_check_descriptors=LocationCheckDescriptors(
                location_type="retract end",
                pipetting_action="dispense",
            ),
            logger=log,
        )
        self._instrument.move_to(
            location=retract_location,
            well_core=self._target_well,
            force_direct=True,
            minimum_z_height=None,
            speed=retract_props.speed,
        )
        retract_delay = retract_props.delay
        if retract_delay.enabled and retract_delay.duration:
            self._instrument.delay(retract_delay.duration)

        blowout_props = retract_props.blowout
        if (
            is_last_retract
            and blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ):
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            self._instrument.blow_out(
                location=retract_location,
                well_core=None,
                in_place=True,
            )
            # A blowout will remove all air gap and liquid (disposal volume) from the tip
            # so delete them from tip state (although practically, there will not be
            # any air gaps in the tip before blowing out in the destination well)
            self._tip_state.delete_last_air_gap_and_liquid()
            self._tip_state.ready_to_aspirate = False

        # A retract will perform total of two air gaps if we need to blow out in source or trash:
        #   - 1st air gap: added before leaving the destination volume to go to src/ trash
        #   - 2nd air gap: added before leaving the blowout location to go to src or tip drop location
        # But if blowout is disabled or is set to Destination well, then only one air gap
        # will be added after retracting, before moving to src or tip drop location.
        # `is_final_air_gap_of_current_retract` tells us whether the next air gap
        # we will be adding, is going to be the last air gap of this step.
        is_final_air_gap_of_current_retract = (
            blowout_props.enabled
            and blowout_props.location == BlowoutLocation.DESTINATION
        ) or not blowout_props.enabled

        # Whether we should add the next air gap depends on the cases as shown below.
        # The main points when deciding this-
        #   - When we have used a conditioning volume, we do not want to add air gaps
        #     while there's still liquid in tip for dispensing
        #   - If we are not using conditioning volume then we want to add gaps just like
        #     we do during the one-to-one transfers
        #   - If this will be the last air gap of the step, if the above two conditions
        #     indicate that we should be adding an air gap, use `add_final_air_gap` as
        #     the final decider of whether to add the air gap.
        if is_final_air_gap_of_current_retract:
            if conditioning_volume > 0:
                add_air_gap = is_last_retract and add_final_air_gap
            else:
                add_air_gap = add_final_air_gap
        else:
            if conditioning_volume > 0:
                add_air_gap = is_last_retract
            else:
                add_air_gap = True

        air_gap_volume = (
            retract_props.air_gap_by_volume.get_for_volume(
                self.tip_state.last_liquid_and_air_gap_in_tip.liquid
            )
            if add_air_gap
            else 0.0
        )

        # Regardless of the blowout location, do touch tip
        # when leaving the dispense well.
        # Add an air gap depending on conditioning volume + whether this is
        # the last step of a multi-dispense sequence + whether this is the last step
        # of the entire liquid distribution.
        self._do_touch_tip_and_air_gap_after_dispense(
            touch_tip_properties=retract_props.touch_tip,
            location=retract_location,
            well=self._target_well,
            air_gap_volume=air_gap_volume,
        )

        if (
            is_last_retract  # We can do a blowout only on the last multi-dispense step
            and blowout_props.enabled
            and blowout_props.location != BlowoutLocation.DESTINATION
        ):
            assert blowout_props.flow_rate is not None
            self._instrument.set_flow_rate(blow_out=blowout_props.flow_rate)
            touch_tip_and_air_gap_location: Union[Location, TrashBin, WasteChute]
            if blowout_props.location == BlowoutLocation.SOURCE:
                if source_location is None or source_well is None:
                    raise RuntimeError(
                        "Blowout location is 'source' but source location &/or well is not provided."
                    )
                # TODO: check if we should add a blowout location z-offset in liq class definition
                self._instrument.blow_out(
                    location=Location(
                        source_well.get_top(0), labware=source_location.labware
                    ),
                    well_core=source_well,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = Location(
                    source_well.get_top(0), labware=source_location.labware
                )
                touch_tip_and_air_gap_well = source_well
            else:
                self._instrument.blow_out(
                    location=trash_location,
                    well_core=None,
                    in_place=False,
                )
                touch_tip_and_air_gap_location = trash_location
                touch_tip_and_air_gap_well = (
                    # We have already established that trash location of `Location` type
                    # has its `labware` as `Well` type.
                    trash_location.labware.as_well()._core  # type: ignore[assignment]
                    if isinstance(trash_location, Location)
                    else None
                )
            # A blowout will remove all air gap and liquid (disposal volume) from the tip
            # so delete them from tip state
            self._tip_state.delete_last_air_gap_and_liquid()
            self._tip_state.ready_to_aspirate = False

            if (
                # Same check as before for when it's the final air gap of current retract
                conditioning_volume > 0
                and is_last_retract
                and add_final_air_gap
            ):
                # The volume in tip at this point should be 0uL
                air_gap_volume = retract_props.air_gap_by_volume.get_for_volume(0)
            else:
                air_gap_volume = 0
            # Do touch tip and air gap again after blowing out into source well or trash
            self._do_touch_tip_and_air_gap_after_dispense(
                touch_tip_properties=retract_props.touch_tip,
                location=touch_tip_and_air_gap_location,
                well=touch_tip_and_air_gap_well,
                air_gap_volume=air_gap_volume,
            )

    def _do_touch_tip_and_air_gap_after_dispense(  # noqa: C901
        self,
        touch_tip_properties: TouchTipProperties,
        location: Union[Location, TrashBin, WasteChute],
        well: Optional[WellCore],
        air_gap_volume: float,
    ) -> None:
        """Perform touch tip and air gap as part of post-dispense retract.

        If the retract location is at or above the safe location of
        AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP, then add the air gap at the retract location
        (where the pipette is already assumed to be at).

        If the retract location is below the safe location, then move to the safe location
        and then add the air gap.

        Note: if the plunger needs to be adjusted to prepare for aspirate, it will be done
        at the same location where the air gap will be added.
        """
        if touch_tip_properties.enabled:
            assert (
                touch_tip_properties.speed is not None
                and touch_tip_properties.z_offset is not None
                and touch_tip_properties.mm_from_edge is not None
            )
            # TODO:, check that when blow out is a non-dest-well,
            #  whether the touch tip params from transfer props should be used for
            #  both dest-well touch tip and non-dest-well touch tip.
            if isinstance(location, Location) and well is not None:
                try:
                    self._instrument.touch_tip(
                        location=location,
                        well_core=well,
                        radius=1,
                        z_offset=touch_tip_properties.z_offset,
                        speed=touch_tip_properties.speed,
                        mm_from_edge=touch_tip_properties.mm_from_edge,
                    )
                except TouchTipDisabledError:
                    # TODO: log a warning
                    pass

                # Move back to the 'retract' position
                self._instrument.move_to(
                    location=location,
                    well_core=well,
                    force_direct=True,
                    minimum_z_height=None,
                    # Full speed because the tip will already be out of the liquid
                    speed=None,
                )
        if air_gap_volume > 0 or not self._tip_state.ready_to_aspirate:
            # If we need to move the plunger up either to prepare for aspirate or to add air gap,
            # move to a safe location above the well if the retract location is not already
            # at or above this safe location
            if isinstance(location, Location):
                assert well is not None  # For mypy purposes only
                if (
                    location.point.z
                    < well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP).z
                ):
                    self._instrument.move_to(
                        location=Location(
                            point=Point(
                                location.point.x,
                                location.point.y,
                                well.get_top(AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP).z,
                            ),
                            labware=location.labware,
                        ),
                        well_core=well,
                        force_direct=True,
                        minimum_z_height=None,
                        speed=None,
                    )
            else:
                if (
                    location.offset.z
                    < location.top(
                        x=0, y=0, z=AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP
                    ).offset.z
                ):
                    self._instrument.move_to(
                        location=location.top(
                            x=location.offset.x,
                            y=location.offset.y,
                            z=AIR_GAP_LOC_Z_OFFSET_FROM_WELL_TOP,
                        ),
                        well_core=None,
                        force_direct=True,
                        minimum_z_height=None,
                        speed=None,
                    )

            if not self._tip_state.ready_to_aspirate:
                self._instrument.prepare_to_aspirate()
                self._tip_state.ready_to_aspirate = True
            if air_gap_volume > 0:
                self._add_air_gap(air_gap_volume=air_gap_volume)

    def _add_air_gap(
        self,
        air_gap_volume: float,
    ) -> None:
        """Add an air gap."""
        if air_gap_volume == 0:
            return
        aspirate_props = self._transfer_properties.aspirate
        correction_volume = aspirate_props.correction_by_volume.get_for_volume(
            self._instrument.get_current_volume() + air_gap_volume
        )
        # The minimum flow rate should be air_gap_volume per second
        flow_rate = max(
            aspirate_props.flow_rate_by_volume.get_for_volume(air_gap_volume),
            air_gap_volume,
        )
        self._instrument.air_gap_in_place(
            volume=air_gap_volume,
            flow_rate=flow_rate,
            correction_volume=correction_volume,
        )
        delay_props = aspirate_props.delay
        if delay_props.enabled and delay_props.duration:
            self._instrument.delay(delay_props.duration)
        self._tip_state.append_air_gap(air_gap_volume)

    def _remove_air_gap(self, location: Union[Location, TrashBin, WasteChute]) -> None:
        """Remove a previously added air gap."""
        last_air_gap = self._tip_state.last_liquid_and_air_gap_in_tip.air_gap
        dispense_props = self._transfer_properties.dispense
        self._instrument.remove_air_gap_during_transfer_with_liquid_class(
            last_air_gap=last_air_gap,
            dispense_props=dispense_props,
            location=location,
        )
        self._tip_state.delete_air_gap(last_air_gap)


def absolute_point_from_position_reference_and_offset(
    well: WellCore,
    well_volume_difference: float,
    position_reference: PositionReference,
    offset: Coordinate,
    mount: Mount,
) -> Point:
    """Return the absolute point, given the well, the position reference and offset.

    If using meniscus as the position reference, well_volume_difference should be specified.
    `well_volume_difference` is the expected *difference* in well volume we want to consider
    when estimating the height of the liquid meniscus after an aspirate/ dispense.
    So, for liquid height estimation after an aspirate, well_volume_difference is
    expected to be a -ve value while for a dispense, it will be a +ve value.
    """
    match position_reference:
        case PositionReference.WELL_TOP:
            reference_point = well.get_top(0)
        case PositionReference.WELL_BOTTOM:
            reference_point = well.get_bottom(0)
        case PositionReference.WELL_CENTER:
            reference_point = well.get_center()
        case PositionReference.LIQUID_MENISCUS:
            estimated_liquid_height = well.estimate_liquid_height_after_pipetting(
                mount=mount,
                operation_volume=well_volume_difference,
            )
            if isinstance(estimated_liquid_height, (float, int)):
                reference_point = well.get_bottom(z_offset=estimated_liquid_height)
            else:
                # If estimated liquid height gives a SimulatedProbeResult then
                # assume meniscus is at well center.
                # Will this cause more harm than good? Is there a better alternative to this?
                reference_point = well.get_center()
        case _:
            raise ValueError(f"Unknown position reference {position_reference}")
    return reference_point + Point(offset.x, offset.y, offset.z)



"""Command models to retrieve a labware from a Flex Stacker."""

from __future__ import annotations
from typing import Literal, TYPE_CHECKING, Any, Union
from typing_extensions import Type

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..flex_stacker.common import FlexStackerStallOrCollisionError
from ...errors import (
    ErrorOccurrence,
    CannotPerformModuleAction,
    LocationIsOccupiedError,
    FlexStackerLabwarePoolNotYetDefinedError,
)
from ...resources import ModelUtils
from ...state import update_types
from ...types import (
    ModuleLocation,
    OnLabwareLocation,
    LabwareLocationSequence,
    LabwareLocation,
    LoadedLabware,
)
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.errors.exceptions import FlexStackerStallError
from opentrons.calibration_storage.helpers import uri_from_details

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.state.module_substates import FlexStackerSubState
    from opentrons.protocol_engine.execution import EquipmentHandler

RetrieveCommandType = Literal["flexStacker/retrieve"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class RetrieveParams(BaseModel):
    """Input parameters for a labware retrieval command."""

    moduleId: str = Field(
        ...,
        description="Unique ID of the Flex Stacker.",
    )
    labwareId: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional ID to assign to this labware. If None, an ID "
        "will be generated.",
        json_schema_extra=_remove_default,
    )
    displayName: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional user-specified display name "
        "or label for this labware.",
        json_schema_extra=_remove_default,
    )
    adapterId: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional ID to assign to an adapter. If None, an ID "
        "will be generated.",
        json_schema_extra=_remove_default,
    )
    lidId: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional ID to assign to a lid. If None, an ID "
        "will be generated.",
        json_schema_extra=_remove_default,
    )


class RetrieveResult(BaseModel):
    """Result data from a labware retrieval command."""

    labwareId: str = Field(
        ...,
        description="The labware ID of the primary retrieved labware.",
    )
    adapterId: str | SkipJsonSchema[None] = Field(
        None,
        description="The optional Adapter Labware ID of the adapter under a primary labware.",
    )
    lidId: str | SkipJsonSchema[None] = Field(
        None,
        description="The optional Lid Labware ID of the lid on a primary labware.",
    )
    primaryLocationSequence: LabwareLocationSequence = Field(
        ..., description="The origin location of the primary labware."
    )
    lidLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = Field(
        None,
        description="The origin location of the adapter labware under a primary labware.",
    )
    adapterLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = Field(
        None, description="The origin location of the lid labware on a primary labware."
    )
    primaryLabwareURI: str = Field(
        ...,
        description="The labware definition URI of the primary labware.",
    )
    adapterLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )


_ExecuteReturn = Union[
    SuccessData[RetrieveResult],
    DefinedErrorData[FlexStackerStallOrCollisionError],
]


class RetrieveImpl(AbstractCommandImpl[RetrieveParams, _ExecuteReturn]):
    """Implementation of a labware retrieval command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    async def _load_labware_from_pool(
        self, params: RetrieveParams, stacker_state: FlexStackerSubState
    ) -> tuple[RetrieveResult, update_types.StateUpdate]:
        state_update = update_types.StateUpdate()

        # If there is an adapter load it
        adapter_lw = None
        lid_lw = None
        definitions_by_id: dict[str, LabwareDefinition] = {}
        offset_ids_by_id: dict[str, str | None] = {}
        display_names_by_id: dict[str, str | None] = {}
        new_locations_by_id: dict[str, LabwareLocation] = {}
        labware_by_id: dict[str, LoadedLabware] = {}
        adapter_uri: str | None = None
        if stacker_state.pool_adapter_definition is not None:
            adapter_location = ModuleLocation(moduleId=params.moduleId)
            adapter_lw = await self._equipment.load_labware_from_definition(
                definition=stacker_state.pool_adapter_definition,
                location=adapter_location,
                labware_id=params.adapterId,
                labware_pending_load=labware_by_id,
            )
            definitions_by_id[adapter_lw.labware_id] = adapter_lw.definition
            offset_ids_by_id[adapter_lw.labware_id] = adapter_lw.offsetId
            display_names_by_id[
                adapter_lw.labware_id
            ] = adapter_lw.definition.metadata.displayName
            new_locations_by_id[adapter_lw.labware_id] = adapter_location
            adapter_uri = str(
                uri_from_details(
                    namespace=adapter_lw.definition.namespace,
                    load_name=adapter_lw.definition.parameters.loadName,
                    version=adapter_lw.definition.version,
                )
            )
            labware_by_id[adapter_lw.labware_id] = LoadedLabware.model_construct(
                id=adapter_lw.labware_id,
                location=adapter_location,
                loadName=adapter_lw.definition.parameters.loadName,
                definitionUri=adapter_uri,
                offsetId=None,
            )
        # Always load the primary labware
        if stacker_state.pool_primary_definition is None:
            raise CannotPerformModuleAction(
                f"Flex Stacker {params.moduleId} has no labware to retrieve"
            )
        primary_location: ModuleLocation | OnLabwareLocation = (
            ModuleLocation(moduleId=params.moduleId)
            if adapter_lw is None
            else OnLabwareLocation(labwareId=adapter_lw.labware_id)
        )
        loaded_labware = await self._equipment.load_labware_from_definition(
            definition=stacker_state.pool_primary_definition,
            location=primary_location,
            labware_id=params.labwareId,
            labware_pending_load={lw_id: lw for lw_id, lw in labware_by_id.items()},
        )
        definitions_by_id[loaded_labware.labware_id] = loaded_labware.definition
        offset_ids_by_id[loaded_labware.labware_id] = loaded_labware.offsetId
        display_names_by_id[
            loaded_labware.labware_id
        ] = loaded_labware.definition.metadata.displayName
        new_locations_by_id[loaded_labware.labware_id] = primary_location
        primary_uri = str(
            uri_from_details(
                namespace=loaded_labware.definition.namespace,
                load_name=loaded_labware.definition.parameters.loadName,
                version=loaded_labware.definition.version,
            )
        )
        labware_by_id[loaded_labware.labware_id] = LoadedLabware.model_construct(
            id=loaded_labware.labware_id,
            location=primary_location,
            loadName=loaded_labware.definition.parameters.loadName,
            definitionUri=primary_uri,
        )

        lid_uri: str | None = None
        # If there is a lid load it
        if stacker_state.pool_lid_definition is not None:
            lid_location = OnLabwareLocation(labwareId=loaded_labware.labware_id)
            lid_lw = await self._equipment.load_labware_from_definition(
                definition=stacker_state.pool_lid_definition,
                location=lid_location,
                labware_id=params.lidId,
                labware_pending_load={lw_id: lw for lw_id, lw in labware_by_id.items()},
            )
            lid_uri = str(
                uri_from_details(
                    namespace=lid_lw.definition.namespace,
                    load_name=lid_lw.definition.parameters.loadName,
                    version=lid_lw.definition.version,
                )
            )
            definitions_by_id[lid_lw.labware_id] = lid_lw.definition
            offset_ids_by_id[lid_lw.labware_id] = lid_lw.offsetId
            display_names_by_id[
                lid_lw.labware_id
            ] = lid_lw.definition.metadata.displayName
            new_locations_by_id[lid_lw.labware_id] = lid_location
            labware_by_id[lid_lw.labware_id] = LoadedLabware.model_construct(
                id=lid_lw.labware_id,
                location=lid_location,
                loadName=lid_lw.definition.parameters.loadName,
                definitionUri=lid_uri,
                offsetId=None,
            )

        # Get the labware dimensions for the labware being retrieved,
        # which is the first one in the hopper labware id list
        primary_location_sequence = (
            self._state_view.geometry.get_predicted_location_sequence(
                new_locations_by_id[loaded_labware.labware_id], labware_by_id
            )
        )
        adapter_location_sequence = (
            self._state_view.geometry.get_predicted_location_sequence(
                new_locations_by_id[adapter_lw.labware_id], labware_by_id
            )
            if adapter_lw is not None
            else None
        )
        lid_location_sequence = (
            self._state_view.geometry.get_predicted_location_sequence(
                new_locations_by_id[lid_lw.labware_id], labware_by_id
            )
            if lid_lw is not None
            else None
        )

        state_update.set_batch_loaded_labware(
            definitions_by_id=definitions_by_id,
            display_names_by_id=display_names_by_id,
            offset_ids_by_id=offset_ids_by_id,
            new_locations_by_id=new_locations_by_id,
        )
        state_update.update_flex_stacker_labware_pool_count(
            module_id=params.moduleId, count=stacker_state.pool_count - 1
        )

        if lid_lw is not None:
            state_update.set_lids(
                parent_labware_ids=[loaded_labware.labware_id],
                lid_ids=[lid_lw.labware_id],
            )

        return (
            RetrieveResult(
                labwareId=loaded_labware.labware_id,
                adapterId=adapter_lw.labware_id if adapter_lw is not None else None,
                lidId=lid_lw.labware_id if lid_lw is not None else None,
                primaryLocationSequence=primary_location_sequence,
                adapterLocationSequence=adapter_location_sequence,
                lidLocationSequence=lid_location_sequence,
                primaryLabwareURI=primary_uri,
                adapterLabwareURI=adapter_uri,
                lidLabwareURI=lid_uri,
            ),
            state_update,
        )

    async def execute(self, params: RetrieveParams) -> _ExecuteReturn:
        """Execute the labware retrieval command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        pool_definitions = stacker_state.get_pool_definition_ordered_list()
        if pool_definitions is None:
            location = self._state_view.modules.get_location(params.moduleId)
            raise FlexStackerLabwarePoolNotYetDefinedError(
                message=f"The Flex Stacker in {location} has not been configured yet and cannot be filled."
            )

        if stacker_state.pool_count == 0:
            raise CannotPerformModuleAction(
                message="Cannot retrieve labware from Flex Stacker because it contains no labware"
            )

        stacker_loc = ModuleLocation(moduleId=params.moduleId)
        # Allow propagation of ModuleNotAttachedError.
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        try:
            self._state_view.labware.raise_if_labware_in_location(stacker_loc)
        except LocationIsOccupiedError:
            raise CannotPerformModuleAction(
                "Cannot retrieve a labware from Flex Stacker if the carriage is occupied"
            )

        retrieve_result, state_update = await self._load_labware_from_pool(
            params, stacker_state
        )

        labware_height = self._state_view.geometry.get_height_of_labware_stack(
            definitions=pool_definitions
        )

        try:
            if stacker_hw is not None:
                await stacker_hw.dispense_labware(labware_height=labware_height)
        except FlexStackerStallError as e:
            return DefinedErrorData(
                public=FlexStackerStallOrCollisionError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                ),
            )

        # Update the state to reflect the labware is now in the Flex Stacker slot
        # todo(chb, 2025-02-19): This ModuleLocation piece should probably instead be an AddressableAreaLocation
        # but that has implications for where labware are set by things like module.load_labware(..) and what
        # happens when we move labware.
        stacker_area = (
            self._state_view.modules.ensure_and_convert_module_fixture_location(
                deck_slot=self._state_view.modules.get_location(
                    params.moduleId
                ).slotName,
                model=self._state_view.modules.get(params.moduleId).model,
            )
        )
        state_update.set_addressable_area_used(stacker_area)

        return SuccessData(
            public=retrieve_result,
            state_update=state_update,
        )


class Retrieve(BaseCommand[RetrieveParams, RetrieveResult, ErrorOccurrence]):
    """A command to retrieve a labware from a Flex Stacker."""

    commandType: RetrieveCommandType = "flexStacker/retrieve"
    params: RetrieveParams
    result: RetrieveResult | None = None

    _ImplementationCls: Type[RetrieveImpl] = RetrieveImpl


class RetrieveCreate(BaseCommandCreate[RetrieveParams]):
    """A request to execute a Flex Stacker retrieve command."""

    commandType: RetrieveCommandType = "flexStacker/retrieve"
    params: RetrieveParams

    _CommandCls: Type[Retrieve] = Retrieve



import logging
from typing import Dict, TYPE_CHECKING

from opentrons.protocol_api import ProtocolContext, InstrumentContext, Labware, Well
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.types import Point, Location
from opentrons_shared_data.protocol.constants import (
    JsonPipetteCommand,
    JsonRobotCommand,
)


if TYPE_CHECKING:
    from opentrons_shared_data.protocol.types import (
        JsonProtocolV3,
        JsonProtocol,
        PipetteAccessParams,
        StandardLiquidHandlingParams,
        DelayParams,
        FlowRateParams,
        TouchTipParams,
        PipetteAccessWithOffsetParams,
        BlowoutParams,
        MoveToSlotParams,
    )
    from opentrons.protocols.execution.dev_types import PipetteDispatch

MODULE_LOG = logging.getLogger(__name__)


def load_pipettes_from_json(
    ctx: ProtocolContext, protocol: "JsonProtocol"
) -> Dict[str, InstrumentContext]:
    pipette_data = protocol["pipettes"]
    pipettes_by_id = {}
    for pipette_id, props in pipette_data.items():
        mount = props["mount"]
        name = props["name"]
        instr = ctx.load_instrument(name, mount)
        pipettes_by_id[pipette_id] = instr

    return pipettes_by_id


def _get_well(
    loaded_labware: Dict[str, Labware], params: "PipetteAccessParams"
) -> Well:
    labwareId = params["labware"]
    well = params["well"]
    plate = loaded_labware[labwareId]
    return plate[well]


# TODO (Ian 2019-04-05) once Pipette commands allow flow rate as an
# absolute value (not % value) as an argument in
# aspirate/dispense/blowout/air_gap fns, remove this
def _set_flow_rate(pipette: InstrumentContext, params: "FlowRateParams") -> None:
    """
    Set flow rate in uL/mm, to value obtained from command's params.
    """
    flow_rate_param = params["flowRate"]

    if not (flow_rate_param > 0):
        raise RuntimeError("Positive flowRate param required")

    pipette.flow_rate.aspirate = flow_rate_param
    pipette.flow_rate.dispense = flow_rate_param
    pipette.flow_rate.blow_out = flow_rate_param


def load_labware_from_json_defs(
    ctx: ProtocolContext, protocol: "JsonProtocolV3"
) -> Dict[str, Labware]:
    protocol_labware = protocol["labware"]
    definitions = protocol["labwareDefinitions"]
    loaded_labware = {}

    for labware_id, props in protocol_labware.items():
        slot = props["slot"]
        definition = definitions[props["definitionId"]]
        label = props.get("displayName", None)
        loaded_labware[labware_id] = ctx.load_labware_from_definition(
            definition, slot, label
        )

    return loaded_labware


def _get_location_with_offset(
    loaded_labware: Dict[str, Labware], params: "PipetteAccessWithOffsetParams"
) -> Location:
    well = _get_well(loaded_labware, params)

    # Never move to the bottom of the fixed trash
    if LabwareLike(well).is_fixed_trash():
        return well.top()

    offset_from_bottom = params["offsetFromBottomMm"]
    bottom = well.bottom()
    return bottom.move(Point(z=offset_from_bottom))


def _delay(context: ProtocolContext, params: "DelayParams") -> None:
    wait = params["wait"]
    message = params.get("message")
    if wait is None or wait is False:
        raise ValueError("Delay must be true, or a number")
    elif wait is True:
        message = message or "Pausing until user resumes"
        context.pause(msg=message)
    else:
        context.delay(seconds=wait, msg=message)


def _blowout(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "BlowoutParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    well = _get_well(loaded_labware, params)
    _set_flow_rate(pipette, params)
    pipette.blow_out(well)


def _pick_up_tip(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "PipetteAccessParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    well = _get_well(loaded_labware, params)
    pipette.pick_up_tip(well)


def _drop_tip(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "PipetteAccessParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    well = _get_well(loaded_labware, params)
    pipette.drop_tip(well)


def _aspirate(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "StandardLiquidHandlingParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    location = _get_location_with_offset(loaded_labware, params)
    volume = params["volume"]
    _set_flow_rate(pipette, params)
    pipette.aspirate(volume, location)


def _dispense(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "StandardLiquidHandlingParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    location = _get_location_with_offset(loaded_labware, params)
    volume = params["volume"]
    _set_flow_rate(pipette, params)
    pipette.dispense(volume, location)


def _air_gap(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "StandardLiquidHandlingParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    offset_from_bottom = params["offsetFromBottomMm"]
    volume = params["volume"]
    _set_flow_rate(pipette, params)
    well = _get_well(loaded_labware, params)
    offset_from_top = offset_from_bottom - well.geometry._depth

    # NOTE(IL, 2020-06-25): air_gap API fn is stateful, uses location
    # cache. The JSON atomic command should be stateless. We'll
    # explicitly move_to the specified well to ensure the location
    # cache is set to that well.
    pipette.move_to(well.top(offset_from_top))
    pipette.air_gap(volume, offset_from_top)


def _touch_tip(
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
    params: "TouchTipParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    location = _get_location_with_offset(loaded_labware, params)
    well = _get_well(loaded_labware, params)
    # convert mmFromBottom to v_offset
    v_offset = location.point.z - well.top().point.z
    pipette.touch_tip(well, v_offset=v_offset)


def _move_to_slot(
    context: ProtocolContext,
    instruments: Dict[str, InstrumentContext],
    params: "MoveToSlotParams",
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]
    slot = params["slot"]
    if slot not in context.deck:
        raise ValueError('Invalid "slot" for "moveToSlot": {}'.format(slot))
    slot_obj = context.deck.position_for(slot)

    offset = params.get("offset", {})
    offsetPoint = Point(offset.get("x", 0), offset.get("y", 0), offset.get("z", 0))

    pipette.move_to(
        slot_obj.move(offsetPoint),
        force_direct=params.get("forceDirect", False),
        minimum_z_height=params.get("minimumZHeight"),
    )


dispatcher_map: "PipetteDispatch" = {
    JsonRobotCommand.delay.value: _delay,
    JsonPipetteCommand.blowout.value: _blowout,
    JsonPipetteCommand.pickUpTip.value: _pick_up_tip,
    JsonPipetteCommand.dropTip.value: _drop_tip,
    JsonPipetteCommand.aspirate.value: _aspirate,
    JsonPipetteCommand.dispense.value: _dispense,
    JsonPipetteCommand.touchTip.value: _touch_tip,
    JsonPipetteCommand.moveToSlot.value: _move_to_slot,
}


def dispatch_json(
    context: ProtocolContext,
    protocol_data: "JsonProtocolV3",
    instruments: Dict[str, InstrumentContext],
    loaded_labware: Dict[str, Labware],
) -> None:
    commands = protocol_data["commands"]

    pipette_commands = {
        JsonPipetteCommand.blowout.value,
        JsonPipetteCommand.pickUpTip.value,
        JsonPipetteCommand.dropTip.value,
        JsonPipetteCommand.aspirate.value,
        JsonPipetteCommand.dispense.value,
        JsonPipetteCommand.touchTip.value,
    }

    for command_item in commands:
        command_type = command_item["command"]
        params = command_item["params"]

        # different `_command` helpers take different args
        if command_type in pipette_commands:
            dispatcher_map[command_type](  # type: ignore[call-arg]
                instruments, loaded_labware, params  # type: ignore[arg-type]
            )
        elif command_type == JsonRobotCommand.delay.value:
            dispatcher_map[command_type](context, params)  # type: ignore
        elif command_type == JsonPipetteCommand.moveToSlot.value:
            dispatcher_map[command_type](context, instruments, params)  # type: ignore
        else:
            raise RuntimeError("Unsupported command type {}".format(command_type))



class DurationEstimatorException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(
            f"Error encountered while estimating protocol duration: '{message}'"
        )



"""Equipment command side-effect logic."""

from dataclasses import dataclass
from typing import Optional, overload, List

from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.types import MountType
from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.modules import (
    AbstractModule,
    MagDeck,
    HeaterShaker,
    TempDeck,
    Thermocycler,
    AbsorbanceReader,
    FlexStacker,
)
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine.state.module_substates import (
    MagneticModuleId,
    HeaterShakerModuleId,
    TemperatureModuleId,
    ThermocyclerModuleId,
    AbsorbanceReaderId,
    FlexStackerId,
)
from ..errors import (
    FailedToLoadPipetteError,
    LabwareDefinitionDoesNotExistError,
    ModuleNotAttachedError,
)
from ..resources import (
    LabwareDataProvider,
    ModuleDataProvider,
    ModelUtils,
    pipette_data_provider,
)
from ..state.state import StateStore
from ..state.modules import HardwareModule
from ..types import (
    LabwareLocation,
    DeckSlotLocation,
    LabwareOffset,
    ModuleModel,
    ModuleDefinition,
    AddressableAreaLocation,
    LoadedLabware,
)


@dataclass(frozen=True)
class LoadedLabwareData:
    """The result of a load labware procedure."""

    labware_id: str
    definition: LabwareDefinition
    offsetId: Optional[str]


@dataclass(frozen=True)
class ReloadedLabwareData:
    """The result of a reload labware procedure."""

    location: LabwareLocation
    offsetId: Optional[str]


@dataclass(frozen=True)
class LoadedPipetteData:
    """The result of a load pipette procedure."""

    pipette_id: str
    serial_number: str
    static_config: pipette_data_provider.LoadedStaticPipetteData


@dataclass(frozen=True)
class LoadedModuleData:
    """The result of a load module procedure."""

    module_id: str
    serial_number: Optional[str]
    definition: ModuleDefinition


@dataclass(frozen=True)
class LoadedConfigureForVolumeData:
    """The result of a load liquid class procedure."""

    pipette_id: str
    serial_number: str
    volume: float
    static_config: pipette_data_provider.LoadedStaticPipetteData


class EquipmentHandler:
    """Implementation logic for labware, pipette, and module loading."""

    _hardware_api: HardwareControlAPI
    _state_store: StateStore
    _labware_data_provider: LabwareDataProvider
    _module_data_provider: ModuleDataProvider
    _model_utils: ModelUtils
    _virtual_pipette_data_provider: pipette_data_provider.VirtualPipetteDataProvider

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_store: StateStore,
        labware_data_provider: Optional[LabwareDataProvider] = None,
        module_data_provider: Optional[ModuleDataProvider] = None,
        model_utils: Optional[ModelUtils] = None,
        virtual_pipette_data_provider: Optional[
            pipette_data_provider.VirtualPipetteDataProvider
        ] = None,
    ) -> None:
        """Initialize an EquipmentHandler instance."""
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._labware_data_provider = labware_data_provider or LabwareDataProvider()
        self._module_data_provider = module_data_provider or ModuleDataProvider()
        self._model_utils = model_utils or ModelUtils()
        self._virtual_pipette_data_provider = (
            virtual_pipette_data_provider
            or pipette_data_provider.VirtualPipetteDataProvider()
        )

    async def load_definition_for_details(
        self, load_name: str, namespace: str, version: int
    ) -> tuple[LabwareDefinition, str]:
        """Load the definition for a labware from the parameters passed for it to a command.

        Args:
            load_name: The labware's load name.
            namespace: The labware's namespace.
            version: The labware's version.

        Returns:
            A tuple of the loaded LabwareDefinition object and its definition URI.
        """
        definition_uri = uri_from_details(
            load_name=load_name,
            namespace=namespace,
            version=version,
        )

        try:
            # Try to use existing definition in state.
            return (
                self._state_store.labware.get_definition_by_uri(definition_uri),
                definition_uri,
            )
        except LabwareDefinitionDoesNotExistError:
            definition = await self._labware_data_provider.get_labware_definition(
                load_name=load_name,
                namespace=namespace,
                version=version,
            )
            return definition, definition_uri

    async def load_labware_from_definition(
        self,
        definition: LabwareDefinition,
        location: LabwareLocation,
        labware_id: Optional[str],
        labware_pending_load: dict[str, LoadedLabware] | None = None,
    ) -> LoadedLabwareData:
        """Load labware from already-found definition."""
        definition_uri = uri_from_details(
            load_name=definition.parameters.loadName,
            namespace=definition.namespace,
            version=definition.version,
        )
        return await self._load_labware_from_def_and_uri(
            definition, definition_uri, location, labware_id, labware_pending_load
        )

    async def _load_labware_from_def_and_uri(
        self,
        definition: LabwareDefinition,
        definition_uri: str,
        location: LabwareLocation,
        labware_id: str | None,
        labware_pending_load: dict[str, LoadedLabware] | None,
    ) -> LoadedLabwareData:
        labware_id = (
            labware_id if labware_id is not None else self._model_utils.generate_id()
        )

        # Allow propagation of ModuleNotLoadedError.
        offset_id = self.find_applicable_labware_offset_id(
            labware_definition_uri=definition_uri,
            labware_location=location,
            labware_pending_load=labware_pending_load,
        )

        return LoadedLabwareData(
            labware_id=labware_id, definition=definition, offsetId=offset_id
        )

    async def load_labware(
        self,
        load_name: str,
        namespace: str,
        version: int,
        location: LabwareLocation,
        labware_id: Optional[str],
    ) -> LoadedLabwareData:
        """Load labware by assigning an identifier and pulling required data.

        Args:
            load_name: The labware's load name.
            namespace: The labware's namespace.
            version: The labware's version.
            location: The deck location at which labware is placed.
            labware_id: An optional identifier to assign the labware. If None, an
                identifier will be generated.

        Raises:
            ModuleNotLoadedError: If `location` references a module ID
                that doesn't point to a valid loaded module.

        Returns:
            A LoadedLabwareData object.
        """
        definition, definition_uri = await self.load_definition_for_details(
            load_name, namespace, version
        )
        return await self._load_labware_from_def_and_uri(
            definition, definition_uri, location, labware_id, None
        )

    async def reload_labware(self, labware_id: str) -> ReloadedLabwareData:
        """Reload an already-loaded labware. This cannot change the labware location.

        Args:
            labware_id: The ID of the already-loaded labware.

        Raises:
            LabwareNotLoadedError: If `labware_id` does not reference a loaded labware.

        """
        location = self._state_store.labware.get_location(labware_id)
        definition_uri = self._state_store.labware.get_definition_uri(labware_id)
        offset_id = self.find_applicable_labware_offset_id(
            labware_definition_uri=definition_uri,
            labware_location=location,
        )

        return ReloadedLabwareData(location=location, offsetId=offset_id)

    async def load_pipette(
        self,
        pipette_name: PipetteNameType,
        mount: MountType,
        pipette_id: Optional[str],
        tip_overlap_version: Optional[str],
    ) -> LoadedPipetteData:
        """Ensure the requested pipette is attached.

        Args:
            pipette_name: The pipette name.
            mount: The mount on which pipette must be attached.
            pipette_id: An optional identifier to assign the pipette. If None, an
                identifier will be generated.
            tip_overlap_version: An optional specifier for the version of tip overlap data to use.
                If None, defaults to v0. Does not need to be format checked - this function does it.

        Returns:
            A LoadedPipetteData object.
        """
        # TODO (spp, 2023-05-10): either raise error if using MountType.EXTENSION in
        #  load pipettes command, or change the mount type used to be a restricted
        #  PipetteMountType which has only pipette mounts and not the extension mount.
        use_virtual_pipettes = self._state_store.config.use_virtual_pipettes

        pipette_name_value = (
            pipette_name.value
            if isinstance(pipette_name, PipetteNameType)
            else pipette_name
        )
        sanitized_overlap_version = (
            pipette_data_provider.validate_and_default_tip_overlap_version(
                tip_overlap_version
            )
        )

        pipette_id = pipette_id or self._model_utils.generate_id()
        if not use_virtual_pipettes:
            cache_request = {mount.to_hw_mount(): pipette_name_value}

            # TODO(mc, 2022-12-09): putting the other pipette in the cache request
            # is only to support protocol analysis, since the hardware simulator
            # does not cache requested virtual instruments. Remove per
            # https://opentrons.atlassian.net/browse/RLIQ-258
            other_mount = mount.other_mount()
            other_pipette = self._state_store.pipettes.get_by_mount(other_mount)
            if other_pipette is not None:
                cache_request[other_mount.to_hw_mount()] = (
                    other_pipette.pipetteName.value
                    if isinstance(other_pipette.pipetteName, PipetteNameType)
                    else other_pipette.pipetteName
                )

            # TODO(mc, 2020-10-18): calling `cache_instruments` mirrors the
            # behavior of protocol_context.load_instrument, and is used here as a
            # pipette existence check
            try:
                await self._hardware_api.cache_instruments(cache_request)
            except RuntimeError as e:
                raise FailedToLoadPipetteError(str(e)) from e

            pipette_dict = self._hardware_api.get_attached_instrument(
                mount.to_hw_mount()
            )

            serial_number = pipette_dict["pipette_id"]
            static_pipette_config = pipette_data_provider.get_pipette_static_config(
                pipette_dict=pipette_dict, tip_overlap_version=sanitized_overlap_version
            )

        else:
            serial_number = self._model_utils.generate_id(prefix="fake-serial-number-")
            static_pipette_config = (
                self._virtual_pipette_data_provider.get_virtual_pipette_static_config(
                    pipette_name=pipette_name_value,
                    pipette_id=pipette_id,
                    tip_overlap_version=sanitized_overlap_version,
                )
            )
        serial = serial_number or ""
        return LoadedPipetteData(
            pipette_id=pipette_id,
            serial_number=serial,
            static_config=static_pipette_config,
        )

    async def load_magnetic_block(
        self,
        model: ModuleModel,
        location: AddressableAreaLocation,
        module_id: Optional[str],
    ) -> LoadedModuleData:
        """Ensure the required magnetic block is attached.

        Args:
            model: The model name of the module.
            location: The deck location of the module
            module_id: Optional ID assigned to the module.
                       If None, an ID will be generated.

        Returns:
            A LoadedModuleData object.

        Raises:
            ModuleAlreadyPresentError: A module of a different type is already
                assigned to the requested location.
        """
        assert ModuleModel.is_magnetic_block(
            model
        ), f"Expected Magnetic block and got {model.name}"
        definition = self._module_data_provider.get_definition(model)
        return LoadedModuleData(
            module_id=self._model_utils.ensure_id(module_id),
            serial_number=None,
            definition=definition,
        )

    async def load_module(
        self,
        model: ModuleModel,
        location: AddressableAreaLocation,
        module_id: Optional[str],
    ) -> LoadedModuleData:
        """Ensure the required module is attached.

        Args:
            model: The model name of the module.
            location: The deck location of the module
            module_id: Optional ID assigned to the module.
                       If None, an ID will be generated.

        Returns:
            A LoadedModuleData object.

        Raises:
            ModuleNotAttachedError: A not-yet-assigned module matching the requested
                parameters could not be found in the attached modules list.
            ModuleAlreadyPresentError: A module of a different type is already
                assigned to the requested location.
        """
        # TODO(mc, 2022-02-09): validate module location given deck definition
        use_virtual_modules = self._state_store.config.use_virtual_modules

        if not use_virtual_modules:
            attached_modules = [
                HardwareModule(
                    serial_number=hw_mod.device_info["serial"],
                    definition=self._module_data_provider.get_definition(
                        ModuleModel(hw_mod.model())
                    ),
                )
                for hw_mod in self._hardware_api.attached_modules
            ]

            serial_number_at_locaiton = self._state_store.geometry._addressable_areas.get_fixture_serial_from_deck_configuration_by_addressable_area(
                addressable_area_name=location.addressableAreaName
            )
            cutout_id = self._state_store.geometry._addressable_areas.get_cutout_id_by_deck_slot_name(
                slot_name=self._state_store.geometry._addressable_areas.get_addressable_area_base_slot(
                    location.addressableAreaName
                )
            )

            attached_module = self._state_store.modules.select_hardware_module_to_load(
                model=model,
                location=cutout_id,
                attached_modules=attached_modules,
                expected_serial_number=serial_number_at_locaiton,
            )

        else:
            attached_module = HardwareModule(
                serial_number=self._model_utils.generate_id(
                    prefix="fake-serial-number-"
                ),
                definition=self._module_data_provider.get_definition(model),
            )

        return LoadedModuleData(
            module_id=self._model_utils.ensure_id(module_id),
            serial_number=attached_module.serial_number,
            definition=attached_module.definition,
        )

    async def load_lids(
        self,
        load_name: str,
        namespace: str,
        version: int,
        location: LabwareLocation,
        quantity: int,
        labware_ids: Optional[List[str]] = None,
    ) -> List[LoadedLabwareData]:
        """Load one or many lid labware by assigning an identifier and pulling required data.

        Args:
            load_name: The lid labware's load name.
            namespace: The lid labware's namespace.
            version: The lid labware's version.
            location: The deck location at which lid(s) will be placed.
            quantity: The quantity of lids to load at a location.
            labware_ids: An optional list of identifiers to assign the labware. If None,
                an identifier will be generated.

        Raises:
            ModuleNotLoadedError: If `location` references a module ID
                that doesn't point to a valid loaded module.

        Returns:
            A list of LoadedLabwareData objects.
        """
        definition_uri = uri_from_details(
            load_name=load_name,
            namespace=namespace,
            version=version,
        )
        try:
            # Try to use existing definition in state.
            definition = self._state_store.labware.get_definition_by_uri(definition_uri)
        except LabwareDefinitionDoesNotExistError:
            definition = await self._labware_data_provider.get_labware_definition(
                load_name=load_name,
                namespace=namespace,
                version=version,
            )

        stack_limit = definition.stackLimit if definition.stackLimit is not None else 1
        if quantity > stack_limit:
            raise ValueError(
                f"Requested quantity {quantity} is greater than the stack limit of {stack_limit} provided by definition for {load_name}."
            )

        is_deck_slot_compatible = (
            True
            if definition.parameters.isDeckSlotCompatible is None
            else definition.parameters.isDeckSlotCompatible
        )

        if isinstance(location, DeckSlotLocation) and not is_deck_slot_compatible:
            raise ValueError(
                f"Lid Labware {load_name} cannot be loaded onto a Deck Slot."
            )

        load_labware_data_list: list[LoadedLabwareData] = []
        ids: list[str] = []
        if labware_ids is not None:
            if len(labware_ids) < quantity:
                raise ValueError(
                    f"Requested quantity {quantity} is greater than the number of labware lid IDs provided for {load_name}."
                )
            ids = labware_ids
        else:
            for i in range(quantity):
                ids.append(self._model_utils.generate_id())
        for id in ids:
            load_labware_data_list.append(
                LoadedLabwareData(
                    labware_id=id,
                    definition=definition,
                    offsetId=None,
                )
            )

        return load_labware_data_list

    async def configure_for_volume(
        self, pipette_id: str, volume: float, tip_overlap_version: Optional[str]
    ) -> LoadedConfigureForVolumeData:
        """Ensure the requested volume can be configured for the given pipette.

        Args:
            pipette_id: The identifier for the pipette.
            volume: The volume to configure the pipette for

        Returns:
            A LoadedConfiguredVolumeData object.
        """
        use_virtual_pipettes = self._state_store.config.use_virtual_pipettes
        sanitized_overlap_version = (
            pipette_data_provider.validate_and_default_tip_overlap_version(
                tip_overlap_version
            )
        )

        if not use_virtual_pipettes:
            mount = self._state_store.pipettes.get_mount(pipette_id).to_hw_mount()

            await self._hardware_api.configure_for_volume(mount, volume)
            pipette_dict = self._hardware_api.get_attached_instrument(mount)

            serial_number = pipette_dict["pipette_id"]
            static_pipette_config = pipette_data_provider.get_pipette_static_config(
                pipette_dict=pipette_dict, tip_overlap_version=sanitized_overlap_version
            )

        else:
            model = self._state_store.pipettes.get_model_name(pipette_id)
            self._virtual_pipette_data_provider.configure_virtual_pipette_for_volume(
                pipette_id, volume, model
            )

            serial_number = self._model_utils.generate_id(prefix="fake-serial-number-")
            static_pipette_config = self._virtual_pipette_data_provider.get_virtual_pipette_static_config_by_model_string(
                pipette_model_string=model,
                pipette_id=pipette_id,
                tip_overlap_version=sanitized_overlap_version,
            )

        return LoadedConfigureForVolumeData(
            pipette_id=pipette_id,
            serial_number=serial_number,
            volume=volume,
            static_config=static_pipette_config,
        )

    async def configure_nozzle_layout(
        self,
        pipette_id: str,
        primary_nozzle: Optional[str] = None,
        front_right_nozzle: Optional[str] = None,
        back_left_nozzle: Optional[str] = None,
    ) -> NozzleMap:
        """Ensure the requested nozzle layout is compatible with the current pipette.

        Args:
            pipette_id: The identifier for the pipette.
            primary_nozzle: The nozzle which will be used as the
            front_right_nozzle
            back_left_nozzle

        Returns:
            A NozzleMap object or None.
        """
        use_virtual_pipettes = self._state_store.config.use_virtual_pipettes

        if not use_virtual_pipettes:
            mount = self._state_store.pipettes.get_mount(pipette_id).to_hw_mount()

            await self._hardware_api.update_nozzle_configuration_for_mount(
                mount,
                back_left_nozzle if back_left_nozzle else primary_nozzle,
                front_right_nozzle if front_right_nozzle else primary_nozzle,
                primary_nozzle if back_left_nozzle else None,
            )
            pipette_dict = self._hardware_api.get_attached_instrument(mount)
            nozzle_map = pipette_dict["current_nozzle_map"]

        else:
            model = self._state_store.pipettes.get_model_name(pipette_id)
            self._virtual_pipette_data_provider.configure_virtual_pipette_nozzle_layout(
                pipette_id,
                model,
                back_left_nozzle if back_left_nozzle else primary_nozzle,
                front_right_nozzle if front_right_nozzle else primary_nozzle,
                primary_nozzle if back_left_nozzle else None,
            )
            nozzle_map = (
                self._virtual_pipette_data_provider.get_nozzle_layout_for_pipette(
                    pipette_id
                )
            )

        return nozzle_map

    @overload
    def get_module_hardware_api(
        self,
        module_id: MagneticModuleId,
    ) -> Optional[MagDeck]:
        ...

    @overload
    def get_module_hardware_api(
        self,
        module_id: HeaterShakerModuleId,
    ) -> Optional[HeaterShaker]:
        ...

    @overload
    def get_module_hardware_api(
        self,
        module_id: TemperatureModuleId,
    ) -> Optional[TempDeck]:
        ...

    @overload
    def get_module_hardware_api(
        self,
        module_id: ThermocyclerModuleId,
    ) -> Optional[Thermocycler]:
        ...

    @overload
    def get_module_hardware_api(
        self,
        module_id: AbsorbanceReaderId,
    ) -> Optional[AbsorbanceReader]:
        ...

    @overload
    def get_module_hardware_api(
        self,
        module_id: FlexStackerId,
    ) -> Optional[FlexStacker]:
        ...

    def get_module_hardware_api(self, module_id: str) -> Optional[AbstractModule]:
        """Get the hardware API for a given module."""
        use_virtual_modules = self._state_store.config.use_virtual_modules
        if use_virtual_modules:
            return None

        attached_modules = self._hardware_api.attached_modules
        serial_number = self._state_store.modules.get_serial_number(module_id)
        for mod in attached_modules:
            if mod.device_info["serial"] == serial_number:
                return mod

        raise ModuleNotAttachedError(
            f'No module attached with serial number "{serial_number}"'
            f' for module ID "{module_id}".'
        )

    def find_applicable_labware_offset_id(
        self,
        labware_definition_uri: str,
        labware_location: LabwareLocation,
        labware_pending_load: dict[str, LoadedLabware] | None = None,
    ) -> str | None:
        """Figure out what offset would apply to a labware in the given location.

        Raises:
            ModuleNotLoadedError: If `labware_location` references a module ID
                that doesn't point to a valid loaded module.

        Returns:
            The ID of the labware offset that will apply,
            or None if no labware offset will apply.
        """
        labware_offset_location = (
            self._state_store.geometry.get_projected_offset_location(
                labware_location, labware_pending_load
            )
        )

        if labware_offset_location is None:
            # No offset for off-deck location.
            # Returning None instead of raising an exception allows loading a labware
            # with 'offDeck' as valid location.
            # Also allows using `moveLabware` with 'offDeck' location.
            return None
        offset = self._state_store.labware.find_applicable_labware_offset(
            definition_uri=labware_definition_uri,
            location=labware_offset_location,
        )
        return self._get_id_from_offset(offset)

    @staticmethod
    def _get_id_from_offset(labware_offset: Optional[LabwareOffset]) -> Optional[str]:
        return None if labware_offset is None else labware_offset.id



# noqa: D100

from __future__ import annotations

import enum
from typing import Optional, Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from opentrons.protocol_engine.commands import (
        Command,
        CommandDefinedErrorData,
    )
    from opentrons.protocol_engine.state.config import Config


class ErrorRecoveryType(enum.Enum):
    """Ways to handle a command failure."""

    FAIL_RUN = enum.auto()
    """Permanently fail the entire run.

    TODO(mm, 2024-03-18): This might be a misnomer because failing the run is not
    a decision that's up to Protocol Engine. It's decided by what the caller supplies
    to `ProtocolEngine.finish()`. For example, a Python protocol can
    theoretically swallow the exception and continue on.
    """

    WAIT_FOR_RECOVERY = enum.auto()
    """Enter interactive error recovery mode."""

    CONTINUE_WITH_ERROR = enum.auto()
    """Continue without interruption, carrying on from whatever error state the failed
    command left the engine in.

    This is like `ProtocolEngine.resume_from_recovery(reconcile_false_positive=False)`.
    """

    ASSUME_FALSE_POSITIVE_AND_CONTINUE = enum.auto()
    """Continue without interruption, acting as if the underlying error was a false positive.

    This is like `ProtocolEngine.resume_from_recovery(reconcile_false_positive=True)`.
    """


class ErrorRecoveryPolicy(Protocol):
    """An interface to decide how to handle a command failure.

    This describes a function that Protocol Engine calls after each command failure,
    with the details of that failure. The implementation should inspect those details
    and return an appropriate `ErrorRecoveryType`.

    Args:
        config: The config of the calling `ProtocolEngine`.
        failed_command: The command that failed, in its final `status=="failed"` state.
        defined_error_data: If the command failed with a defined error, details about
            that error. If the command failed with an undefined error, `None`.
            By design, this callable isn't given details about undefined errors,
            since it would be fragile to rely on them.
    """

    @staticmethod
    def __call__(  # noqa: D102
        config: Config,
        failed_command: Command,
        defined_error_data: Optional[CommandDefinedErrorData],
    ) -> ErrorRecoveryType:
        ...


def never_recover(
    config: Config,
    failed_command: Command,
    defined_error_data: Optional[CommandDefinedErrorData],
) -> ErrorRecoveryType:
    """An error recovery policy where error recovery is never attempted.

    This makes sense for things like the `opentrons_simulate` and `opentrons_execute`
    CLIs. Those don't expose any way to bring the run out of recovery mode after it's
    been entered, so we need to avoid entering recovery mode in the first place.
    """
    return ErrorRecoveryType.FAIL_RUN



from typing_extensions import Final
import enum


class NozzleLayout(enum.Enum):
    COLUMN = "COLUMN"
    PARTIAL_COLUMN = "PARTIAL_COLUMN"
    SINGLE = "SINGLE"
    ROW = "ROW"
    QUADRANT = "QUADRANT"
    ALL = "ALL"


COLUMN: Final = NozzleLayout.COLUMN
PARTIAL_COLUMN: Final = NozzleLayout.PARTIAL_COLUMN
SINGLE: Final = NozzleLayout.SINGLE
ROW: Final = NozzleLayout.ROW
ALL: Final = NozzleLayout.ALL

# Set __doc__ manually as a workaround. When this docstring is written the normal way, right after
# the constant definition, Sphinx has trouble picking it up.
COLUMN.__doc__ = """\
A special nozzle configuration type indicating a full single column pick up. Predominantly meant for the 96 channel.

See <ADD REFERENCE HERE> for details on using ``COLUMN`` with :py:obj:`InstrumentContext.configure_nozzle_layout()`.
"""
ALL.__doc__ = """\
A special nozzle configuration type indicating a reset back to default where the pipette will pick up its max capacity of tips.

See <ADD REFERENCE HERE> for details on using ``ALL`` with :py:obj:`InstrumentContext.configure_nozzle_layout()`.
"""



"""Load labware command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal, TypeGuard

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ..errors import LabwareIsNotAllowedInLocationError
from ..resources import labware_validation, fixture_validation
from ..types import (
    LoadableLabwareLocation,
    ModuleLocation,
    ModuleModel,
    OnLabwareLocation,
    DeckSlotLocation,
    AddressableAreaLocation,
    LoadedModule,
)

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from .labware_handling_common import LabwarePositionResultMixin
from ..errors.error_occurrence import ErrorOccurrence
from ..state.update_types import StateUpdate

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import EquipmentHandler


LoadLabwareCommandType = Literal["loadLabware"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class LoadLabwareParams(BaseModel):
    """Payload required to load a labware into a slot."""

    location: LoadableLabwareLocation = Field(
        ...,
        description="Location the labware should be loaded into.",
    )
    loadName: str = Field(
        ...,
        description="Name used to reference a labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The labware definition version.",
    )
    labwareId: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional ID to assign to this labware. If None, an ID "
        "will be generated.",
        json_schema_extra=_remove_default,
    )
    displayName: str | SkipJsonSchema[None] = Field(
        None,
        description="An optional user-specified display name "
        "or label for this labware.",
        # NOTE: v4/5 JSON protocols will always have a displayName which will be the
        #  user-specified label OR the displayName property of the labware's definition.
        # TODO: Make sure v6 JSON protocols don't do that.
        json_schema_extra=_remove_default,
    )


class LoadLabwareResult(LabwarePositionResultMixin):
    """Result data from the execution of a LoadLabware command."""

    definition: LabwareDefinition = Field(
        ...,
        description="The full definition data for this labware.",
    )


class LoadLabwareImplementation(
    AbstractCommandImpl[LoadLabwareParams, SuccessData[LoadLabwareResult]]
):
    """Load labware command implementation."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    def _is_loading_to_module(
        self, location: LoadableLabwareLocation, module_model: ModuleModel
    ) -> TypeGuard[ModuleLocation]:
        if not isinstance(location, ModuleLocation):
            return False

        module: LoadedModule = self._state_view.modules.get(location.moduleId)
        return module.model == module_model

    async def execute(
        self, params: LoadLabwareParams
    ) -> SuccessData[LoadLabwareResult]:
        """Load definition and calibration data necessary for a labware."""
        state_update = StateUpdate()

        # TODO (tz, 8-15-2023): extend column validation to column 1 when working
        # on https://opentrons.atlassian.net/browse/RSS-258 and completing
        # https://opentrons.atlassian.net/browse/RSS-255
        if (
            labware_validation.is_flex_trash(params.loadName)
            and isinstance(params.location, DeckSlotLocation)
            and self._state_view.geometry.get_slot_column(params.location.slotName) != 3
        ):
            raise LabwareIsNotAllowedInLocationError(
                f"{params.loadName} is not allowed in slot {params.location.slotName}"
            )

        if isinstance(params.location, AddressableAreaLocation):
            area_name = params.location.addressableAreaName
            if not (
                fixture_validation.is_deck_slot(params.location.addressableAreaName)
                or fixture_validation.is_abs_reader(params.location.addressableAreaName)
            ):
                raise LabwareIsNotAllowedInLocationError(
                    f"Cannot load {params.loadName} onto addressable area {area_name}"
                )
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                area_name
            )
            state_update.set_addressable_area_used(area_name)
        elif isinstance(params.location, DeckSlotLocation):
            self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
                params.location.slotName.id
            )
            state_update.set_addressable_area_used(params.location.slotName.id)

        verified_location = self._state_view.geometry.ensure_location_not_occupied(
            params.location
        )

        loaded_labware = await self._equipment.load_labware(
            load_name=params.loadName,
            namespace=params.namespace,
            version=params.version,
            location=verified_location,
            labware_id=params.labwareId,
        )

        state_update.set_loaded_labware(
            labware_id=loaded_labware.labware_id,
            offset_id=loaded_labware.offsetId,
            definition=loaded_labware.definition,
            location=verified_location,
            display_name=params.displayName,
        )

        if isinstance(verified_location, OnLabwareLocation):
            self._state_view.labware.raise_if_labware_cannot_be_stacked(
                top_labware_definition=loaded_labware.definition,
                bottom_labware_id=verified_location.labwareId,
            )

        # Validate labware for the absorbance reader
        if self._is_loading_to_module(
            params.location, ModuleModel.ABSORBANCE_READER_V1
        ):
            self._state_view.labware.raise_if_labware_incompatible_with_plate_reader(
                loaded_labware.definition
            )

        self._state_view.labware.raise_if_labware_cannot_be_ondeck(
            location=params.location, labware_definition=loaded_labware.definition
        )

        return SuccessData(
            public=LoadLabwareResult(
                labwareId=loaded_labware.labware_id,
                definition=loaded_labware.definition,
                offsetId=loaded_labware.offsetId,
                locationSequence=self._state_view.geometry.get_predicted_location_sequence(
                    verified_location,
                ),
            ),
            state_update=state_update,
        )


class LoadLabware(BaseCommand[LoadLabwareParams, LoadLabwareResult, ErrorOccurrence]):
    """Load labware command resource model."""

    commandType: LoadLabwareCommandType = "loadLabware"
    params: LoadLabwareParams
    result: Optional[LoadLabwareResult] = None

    _ImplementationCls: Type[LoadLabwareImplementation] = LoadLabwareImplementation


class LoadLabwareCreate(BaseCommandCreate[LoadLabwareParams]):
    """Load labware command creation request."""

    commandType: LoadLabwareCommandType = "loadLabware"
    params: LoadLabwareParams

    _CommandCls: Type[LoadLabware] = LoadLabware



"""Generate a JSON schema against which all create commands statically validate."""

import json
import argparse
import sys
from opentrons.protocol_engine.commands.command_unions import CommandCreateAdapter

from opentrons_shared_data.command import get_newest_schema_version
from opentrons_shared_data.load import get_shared_data_root


def generate_command_schema(version: str) -> str:
    """Generate a JSON Schema that all valid create commands can validate against."""
    schema_as_dict = CommandCreateAdapter.json_schema(mode="validation", by_alias=False)
    schema_as_dict["$id"] = f"opentronsCommandSchemaV{version}"
    schema_as_dict["$schema"] = "http://json-schema.org/draft-07/schema#"
    return json.dumps(schema_as_dict, indent=2, sort_keys=True)


def write_command_schema(json_string: str, version: str) -> None:
    """Write a JSON command schema to the shared-data command schema directory."""
    path = get_shared_data_root() / "command" / "schemas" / f"{version}.json"
    with open(path, "w") as schema_file:
        schema_file.write(json_string)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="generate_command_schema",
        description="Generate A JSON-schema of all possible Create-Commands accepted by the current Protocol Engine",
    )
    parser.add_argument(
        "version",
        type=str,
        nargs="?",
        help="The command schema version. This is a single integer (e.g. 7) that will be used to name the generated"
        " schema file. If not included, it will automatically use the latest version in shared-data.",
    )
    parser.add_argument(
        "--overwrite-shared-data",
        action="store_true",
        help="If used, overwrites the specified or automatically chosen command schema version in shared-data."
        " If not included, the generated schema will be printed to stdout.",
    )
    args = parser.parse_args()

    if args.version is None:
        version_string = get_newest_schema_version()
    else:
        version_string = args.version

    command_schema = generate_command_schema(version_string)

    if args.overwrite_shared_data:
        write_command_schema(command_schema, version_string)
    else:
        print(command_schema)

    sys.exit()

__all__ = ["generate_command_schema"]
